import pytest
from timeit import default_timer as timer

from netgen.geom2d import SplineGeometry
from ngsolve import *
from xfem import *
from xfem import ngsxfemglobals
from xfem import IntegrateX
from xfem import Integrate
from xfem.lsetcurv import *

import numpy as np
from numpy import sqrt

# compare simd to normal Integrate
@pytest.mark.parametrize("maxh", [2**(-k) for k in range(5)])
@pytest.mark.parametrize("element_wise", [True, False])
@pytest.mark.parametrize("order", [4])
def test_simd_integrate(maxh, element_wise, order):
    square = SplineGeometry()
    square.AddRectangle((-1, -1), (1, 1), bc=1)
    ngmesh = square.GenerateMesh(maxh=maxh, quad_dominated=False)
    mesh = Mesh(ngmesh)
    
    levelset = CoefficientFunction(x**2 + y**2) - 0.5
    
    order = 1
    
    V = H1(mesh, order = order, autoupdate=True)
    lset_approx = GridFunction(V)
    InterpolateToP1(levelset, lset_approx)
    
    f = CoefficientFunction(1)
    
    dict = {
        "levelset": levelset,
        "domain_type": NEG,
        "order": order,
        "quad_dir_policy": FIRST
    }
    
    t_normal = 0
    t_simd = 0
    solutions = []
    testruns = 5
    for i in range(testruns):
        # simd disabled
        ngsxfemglobals.simd_eval = False
        start = timer()
        integral_normal = Integrate(f, mesh, VOL)
        end = timer()
        
        t_normal += (end-start)
        
        # simd enabled
        ngsxfemglobals.simd_eval = True
        
        start = timer()
        integral_simd = Integrate(f, mesh, VOL)
        end = timer()
        
        t_simd += (end-start)
        
        solutions.append(integral_simd - integral_normal)
    print()
    print("Time for Integrate (normal): " + str(t_normal/testruns))
    print("Time for Integrate (simd): " + str(t_simd/testruns))
    print("performance ratio (normal/simd): " + str(t_normal/t_simd))
    assert(np.max(solutions) < 1e-8)



# compare simd to normal IntegrateX
@pytest.mark.parametrize("maxh", [2**(-k) for k in range(5)])
@pytest.mark.parametrize("element_wise", [True, False])
@pytest.mark.parametrize("order", [4])
def test_simd_integrateX(maxh, element_wise, order):
    square = SplineGeometry()
    square.AddRectangle((-1, -1), (1, 1), bc=1)
    ngmesh = square.GenerateMesh(maxh=maxh, quad_dominated=False)
    mesh = Mesh(ngmesh)
    
    levelset = CoefficientFunction(x**2 + y**2) - 0.5
    
    V = H1(mesh, order = order, autoupdate=True)
    lset_approx = GridFunction(V)
    InterpolateToP1(levelset, lset_approx)
    
    dict = {
        "levelset": lset_approx,
        "mesh": mesh,
        "order": order,
        "domain_type": NEG
    }
    
    f = CoefficientFunction(1)
    
    t_normal = 0
    t_simd = 0
    solutions = []
    testruns = 5
    for i in range(testruns):
        # simd disabled
        ngsxfemglobals.simd_eval = False
        start = timer()
        integral_normal = IntegrateX(dict, mesh, f)
        end = timer()
        
        t_normal += (end-start)
        
        # simd enabled
        ngsxfemglobals.simd_eval = True
        
        start = timer()
        integral_simd = IntegrateX(dict, mesh, f)
        end = timer()
        
        t_simd += (end-start)
        solutions.append(integral_simd - integral_normal)
    print()
    print("Time for IntegrateX (normal): " + str(t_normal/testruns))
    print("Time for IntegrateX (simd): " + str(t_simd/testruns))
    print("performance ratio (normal/simd): " + str(t_normal/t_simd))
    assert(np.max(solutions) < 1e-7)


test_simd_integrate(0.1, False, 4)
test_simd_integrateX(0.1, False, 4)


from ngsolve import *
from time import sleep
from netgen.geom2d import unit_square
from netgen.geom2d import SplineGeometry
from netgen.meshing import MeshingParameters
from ngsolve.internal import *
from xfem import *
from numpy import pi
from xfem.lset_spacetime import *


square = SplineGeometry()
square.AddRectangle([0,0],[2,2],bc=1)
maxh = 0.3
mesh = Mesh (square.GenerateMesh(maxh=maxh, quad_dominated=False))


h = specialcf.mesh_size

# FE Spaces
k_t = 1
k_s = 1
fes1 = H1(mesh,order=k_s,dirichlet=[1,2,3,4])

lset_order_time = 2
tfe = ScalarTimeFE(k_t) 

st_fes = SpaceTimeFESpace(fes1,tfe)
#visoptions.deformation = 1

tend = 1
delta_t = 1/32
tnew = 0

told = Parameter(0)
tref = ReferenceTimeVariable()
t = told + delta_t*tref

lset_adap_st = LevelSetMeshAdaptation_Spacetime(mesh, order_space = k_s, order_time = lset_order_time,
                                         threshold=0.1, discontinuous_qn=True)

# Problem data

problems =  {"planar": ( CoefficientFunction(0.25*t), CoefficientFunction(0.25),
                         CoefficientFunction(1),CoefficientFunction(0),CoefficientFunction(0)),
             "nonlinear": ( CoefficientFunction((0.25/pi)*sin(2*pi*t)),CoefficientFunction(0.5*cos(2*pi*t)),
                            CoefficientFunction(7/8 + (1/4)*y*y*(2-y)*(2-y)),CoefficientFunction( y*(1-y)*(2-y) ),CoefficientFunction( (1-y)*(2-y) - y*(2-y) - y*(1-y))) }
          
rho,d_rho,q_coeff,d_q_coeff,dd_q_coeff = problems["planar"]

r11 = x - q_coeff - rho

D_const = 2/3
levelset= CoefficientFunction(sqrt(r11*r11) - 0.5*D_const)

fes_lset_slice = H1(mesh, order=1, dirichlet=[])
lset_p1_slice = GridFunction(fes_lset_slice)

t_old = 0
ci = CutInfo(mesh)

while tend - t_old > delta_t/2:
    
    dfm = lset_adap_st.CalcDeformation(levelset,told,t_old,delta_t,calc_kappa = True) 
    lset_p1 = lset_adap_st.lset_p1  
    ci.Update(lset_p1,2)
    print("time = {0}".format(t_old))
    hasneg= BitArray(ci.GetElementsOfType(NEG,VOL))
    hasneg_BND = BitArray(ci.GetElementsOfType(NEG,BND))
    print("hasneg_BND:")
    print(hasneg_BND)
    haspos= BitArray(ci.GetElementsOfType(POS,VOL))
    haspos_BND = BitArray(ci.GetElementsOfType(POS,BND))
    print("haspos_BND:")
    print(haspos_BND)
    hasif= BitArray(ci.GetElementsOfType(IF,VOL))
    hasif_BND = BitArray(ci.GetElementsOfType(IF,BND))
    print("hasif_BND:")
    print(hasif_BND)
    Draw(BitArrayCF(hasneg),mesh,"NEG")
    Draw(BitArrayCF(haspos),mesh,"POS")
    Draw(BitArrayCF(hasif),mesh,"IF")
    lset_p1_slice.vec[:] = lset_p1.vec[0:fes_lset_slice.ndof]
    Draw(IfPos(-lset_p1_slice,CoefficientFunction(1),CoefficientFunction(0)),mesh,"lsetp1")
    input("Continue")
    
    t_old = t_old + delta_t
    told.Set(t_old)


import pytest
from ngsolve import *
from xfem import *
from ngsolve.meshes import *
from netgen.geom2d import SplineGeometry
from netgen.csg import CSGeometry, OrthoBrick, Pnt
from math import pi
from xfem.lsetcurv import *


@pytest.mark.parametrize("integration_quadtrig_newold", [(False, True),(False, False),(True, True)])
@pytest.mark.parametrize("order", [1,2,3])

def test_nxfem(integration_quadtrig_newold,order):
    quad, newintegration = integration_quadtrig_newold
    if quad:
        mesh = MakeStructured2DMesh(quads = True, mapping = lambda x,y: (3*x-1.5,3*y-1.5), nx=15, ny=15)
    else:
        square = SplineGeometry()
        square.AddRectangle([-1.5,-1.5],[1.5,1.5],bc=1)
        mesh = Mesh (square.GenerateMesh(maxh=0.2))
    r44 = (x*x*x*x+y*y*y*y)
    r41 = sqrt(sqrt(x*x*x*x+y*y*y*y))
    r4m3 = (1.0/(r41*r41*r41))
    r66 = (x*x*x*x*x*x+y*y*y*y*y*y)
    r63 = sqrt(r66)
    r22 = (x*x+y*y)
    r21 = sqrt(r22)
    
    levelset = (sqrt(sqrt(x*x*x*x+y*y*y*y)) - 1.0)

    solution = [1.0+pi/2.0-sqrt(2.0)*cos(pi/4.0*r44),pi/2.0*r41]
    alpha = [1.0,2.0]
    coef_f = [ (-1.0*sqrt(2.0)*pi*(pi*cos(pi/4*(r44))*(r66)+3*sin(pi/4*(r44))*(r22))),
              (-2.0*pi*3/2*(r4m3)*(-(r66)/(r44)+(r22))) ]
    
    order = order

    lsetmeshadap = LevelSetMeshAdaptation(mesh, order=order, threshold=1000, discontinuous_qn=True)
    deformation = lsetmeshadap.CalcDeformation(levelset)
    if (newintegration):
        lsetp1 = lsetmeshadap.lset_p1
        gradlsetp1 = grad(lsetp1)
    else:
        lsetp1 = 1.0*lsetmeshadap.lset_p1
        gradlsetp1 = 1.0*grad(lsetmeshadap.lset_p1)

    
    # extended FESpace 
    
    Vh = H1(mesh, order=order, dirichlet=[1,2,3,4])
    Vhx = XFESpace(Vh, lsetp1)
    VhG = FESpace([Vh,Vhx])
    print("unknowns in extended FESpace:", VhG.ndof)
    
    # coefficients / parameters: 
    
    n = 1.0/gradlsetp1.Norm() * gradlsetp1
    h = specialcf.mesh_size
    
    kappa = [CutRatioGF(Vhx.GetCutInfo()),1.0-CutRatioGF(Vhx.GetCutInfo())]
    
    stab = 10*(alpha[1]+alpha[0])*(order+1)*order/h
    
    # expressions of test and trial functions:
    
    u_std, u_x = VhG.TrialFunction()
    v_std, v_x = VhG.TestFunction()
    
    u = [u_std + op(u_x) for op in [neg,pos]]
    v = [v_std + op(v_x) for op in [neg,pos]]
    
    gradu = [grad(u_std) + op(u_x) for op in [neg_grad,pos_grad]]
    gradv = [grad(v_std) + op(v_x) for op in [neg_grad,pos_grad]]
    
    average_flux_u = sum([- kappa[i] * alpha[i] * gradu[i] * n for i in [0,1]])
    average_flux_v = sum([- kappa[i] * alpha[i] * gradv[i] * n for i in [0,1]])
    
    # integration domains (and integration parameter "subdivlvl" and "force_intorder")
    
    lset_neg = { "levelset" : lsetp1, "domain_type" : NEG, "subdivlvl" : 0}
    lset_pos = { "levelset" : lsetp1, "domain_type" : POS, "subdivlvl" : 0}
    lset_if  = { "levelset" : lsetp1, "domain_type" : IF , "subdivlvl" : 0}
    
    # bilinear forms:
    
    a = BilinearForm(VhG, symmetric = True, flags = { })
    a += SymbolicBFI(levelset_domain = lset_neg, form = alpha[0] * gradu[0] * gradv[0])
    a += SymbolicBFI(levelset_domain = lset_pos, form = alpha[1] * gradu[1] * gradv[1])
    a += SymbolicBFI(levelset_domain = lset_if , form =     average_flux_u * (v[0]-v[1]))
    a += SymbolicBFI(levelset_domain = lset_if , form =     average_flux_v * (u[0]-u[1]))
    a += SymbolicBFI(levelset_domain = lset_if , form = stab * (u[0]-u[1]) * (v[0]-v[1]))
    
    f = LinearForm(VhG)
    f += SymbolicLFI(levelset_domain = lset_neg, form = coef_f[0] * v[0])
    f += SymbolicLFI(levelset_domain = lset_pos, form = coef_f[1] * v[1])
    
    gfu = GridFunction(VhG)
    
    gfu.components[0].Set(solution[1], BND)
    
    mesh.SetDeformation(deformation)
    
    a.Assemble();
    f.Assemble();
    rhs = gfu.vec.CreateVector()
    rhs.data = f.vec - a.mat * gfu.vec
    update = gfu.vec.CreateVector()
    update.data = a.mat.Inverse(VhG.FreeDofs()) * rhs;
    gfu.vec.data += update
    
    sol_coef = IfPos(lsetp1,solution[1],solution[0])
    u_coef = gfu.components[0] + IfPos(lsetp1, pos(gfu.components[1]), neg(gfu.components[1]))
    u = [gfu.components[0] + op(gfu.components[1]) for op in [neg,pos]]
    
    err_sqr_coefs = [ (u[i] - solution[i])*(u[i] - solution[i]) for i in [0,1] ]
    
    l2error = sqrt(   Integrate(levelset_domain = lset_neg, cf=err_sqr_coefs[0], mesh=mesh, order=order, heapsize=1000000)
                    + Integrate(levelset_domain = lset_pos, cf=err_sqr_coefs[1], mesh=mesh, order=order, heapsize=1000000) )

    print("L2 error : ",l2error)

    mesh.UnsetDeformation()
    
    if (order == 1):
        assert l2error < 0.06
    if (order == 2):
        assert l2error < 0.004
    if (order == 3):
        assert l2error < 0.0004

if __name__ == "__main__":
    test_nxfem((False,False),1)
    test_nxfem((True,True),2)
    test_nxfem((False,True),3)
    



"""
example of a triangle described by three level set functions
"""

# the constant pi
from math import pi
# ngsolve stuff
from ngsolve import *
# basic xfem functionality
from xfem import *
# basic geometry features (for the background mesh)
from netgen.geom2d import SplineGeometry

# We generate the background mesh of the domain and use a simplicial triangulation
# To obtain a mesh with quadrilaterals use 'quad_dominated=True'

square = SplineGeometry()
square.AddRectangle([-1,-0.5], [1,1.5], bc=1)
mesh = Mesh (square.GenerateMesh(maxh=1, quad_dominated=False))

lsetp1_a = GridFunction(H1(mesh,order=1))
lsetp1_b = GridFunction(H1(mesh,order=1))
lsetp1_c = GridFunction(H1(mesh,order=1))

InterpolateToP1(     y-1,lsetp1_a)
InterpolateToP1( 2*x-y  ,lsetp1_b)
InterpolateToP1(-2*x-y  ,lsetp1_c)

Draw (lsetp1_a, mesh, "lset_a")
Draw (lsetp1_b, mesh, "lset_b")
Draw (lsetp1_c, mesh, "lset_c")

Draw (IfPos(lsetp1_a,0,1) * IfPos(lsetp1_b,0,1) * IfPos(lsetp1_c,0,1), mesh, "lset_mult")

area0 = Integrate(levelset_domain={"levelset" : lsetp1_a, "domain_type": NEG},
                  mesh=mesh, cf=1, order=0)
print("area0 = {:10.8f}".format(area0))                     
print("area0 error = {:4.3e}".format(abs(area0-3)))
assert abs(area0-3) < 1e-12

area = Integrate(levelset_domain={"levelset" : (lsetp1_a,lsetp1_b,lsetp1_c), "domain_type": (NEG,NEG,NEG)},
                         mesh=mesh, cf=1, order=0)
print("area = {:10.8f}".format(area))                     
print("area error = {:4.3e}".format(abs(area-0.5)))  
assert abs(area-0.5) < 1e-12

area_invert = Integrate(levelset_domain={"levelset" : (lsetp1_a,lsetp1_b,lsetp1_c), "domain_type":
                                         [(NEG,NEG,POS), (NEG,POS,NEG), (NEG,POS,POS), (POS,NEG,NEG),
                                          (POS,NEG,POS), (POS,POS,NEG), (POS,POS,POS),
                                         ]},
                         mesh=mesh, cf=1, order=0)
print("area_invert = {:10.8f}".format(area_invert))                     
print("area_invert error = {:4.3e}".format(abs(area_invert-3.5)))  
assert abs(area_invert-3.5) < 1e-12
                   

area = Integrate(levelset_domain={"levelset" : (lsetp1_a,lsetp1_b,lsetp1_c), "domain_type": [(NEG,NEG,NEG)]},
                         mesh=mesh, cf=1, order=0)
print("[duplicate] area = {:10.8f}".format(area))                     
print("[duplicate] area error = {:4.3e}".format(abs(area-0.5)))  
assert abs(area-0.5) < 1e-12
                   
length1 = Integrate(levelset_domain={"levelset" : (lsetp1_a,lsetp1_b,lsetp1_c), "domain_type": (IF,NEG,NEG)},
                         mesh=mesh, cf=1, order=0)
print("length1 =", length1)                     
print("length1 error =", abs(length1-1))                     
assert abs(length1-1) < 1e-12

length2 = Integrate(levelset_domain={"levelset" : (lsetp1_a,lsetp1_b,lsetp1_c), "domain_type": (NEG,IF,NEG)},
                         mesh=mesh, cf=1, order=0)
print("length2 =", length2)                     
print("length2 error =", abs(length2-sqrt(5/4)))                     
assert abs(length2-sqrt(5/4)) < 1e-12

length3 = Integrate(levelset_domain={"levelset" : (lsetp1_a,lsetp1_b,lsetp1_c), "domain_type": (NEG,NEG,IF)},
                         mesh=mesh, cf=1, order=0)
print("length3 =", length3)                     
print("length3 error =", abs(length3-sqrt(5/4)))                     
assert abs(length3-sqrt(5/4)) < 1e-12

point_val_y = Integrate(levelset_domain={"levelset" : (lsetp1_a,lsetp1_b,lsetp1_c), "domain_type": (NEG,IF,IF)},
                        mesh=mesh, cf=y, order=0)
print("point_val_y =", point_val_y)                     
print("point_val_y error =", abs(point_val_y-0))
assert abs(point_val_y) < 1e-12


point_val_yp = Integrate(levelset_domain={"levelset" : (lsetp1_a,lsetp1_b,lsetp1_c), "domain_type": (NEG,IF,IF)},
                         mesh=mesh, cf=1-y, order=0)
print("point_val_yp =", point_val_yp)                     
print("point_val_yp error =", abs(point_val_yp-1))
assert abs(point_val_yp-1) < 1e-12


# inner = { "levelsets" : (lsetp1_a,lsetp1_b,lsetp1_c),
#           "domain_type" : (NEG,NEG,NEG)}

# boundary = { "levelsets" : (lsetp1_a,lsetp1_b,lsetp1_c),
#              "domain_type" : (IF,NEG,NEG) | (NEG,IF,NEG) | (NEG,NEG,IF)}

# outer = { "levelsets" : (lsetp1_a,lsetp1_b,lsetp1_c),
#           "domain_type" : ~(NEG,NEG,NEG)}

# triangle_area = Integrate( levelset_domain=inner, cf=1, mesh=mesh, order=2)
# triangle_boundary_length = Integrate( levelset_domain=boundary, cf=1, mesh=mesh, order=2)
# outer_area = Integrate( levelset_domain=outer, cf=1, mesh=mesh, order=2)



"""
In this example we solve a scalar Laplace-Beltrami problem with a
similar discretisation method to the one used in tracefem.py. However,
we use a 3D (background mesh dimension) problem and higher order method
this time. To be robust w.r.t. the interface position also in the
condition number we use the normal diffusion stabilization, cf.[1,2].

Used features:
--------------
* Higher order geometry approximation, cf. jupyter tutorial `lsetint`

* Restricted finite element space to condense the system to active dofs,
  cf. jupyter tutorial `basics`

* Visualization: The visualization of the solution is most convenient
  with paraview and the generated vtk file.

Literature:
-----------
[1] J. Grande, C. Lehrenfeld, A. Reusken, Analysis of a high-order trace
    finite element method for PDEs on level set surfaces.
    SIAM Journal on Numerical Analysis, 56(1):228-255, 2018.
[2] E. Burman, P. Hansbo, M. G. Larson, A. Massing, Cut  finite  element
    methods  for  partial  differential  equations  on  embedded  manifolds
    of  arbitrary  codimensions.  ESAIM: M2AN, 52(6):2247-2282, 2018.
"""

# ------------------------------ LOAD LIBRARIES -------------------------------
import sys
from netgen.csg import CSGeometry, OrthoBrick, Pnt
from ngsolve import *
from ngsolve.internal import *
from xfem import *
from xfem.lsetcurv import *
from math import pi

# -------------------------------- PARAMETERS ---------------------------------
# Mesh diameter
maxh = 0.6
# Bisect cut elements of (initial) mesh
n_cut_ref = 2
# Polynomial order of FE space
order = 2

# Problem parameters
reac_cf = 1
diff_cf = 1

# Geometry
cube = CSGeometry()
cube.Add(OrthoBrick(Pnt(-1.41, -1.41, -1.41), Pnt(1.41, 1.41, 1.41)))
mesh = Mesh(cube.GenerateMesh(maxh=maxh, quad_dominated=False))

levelset = sqrt(x**2 + y**2 + z**2) - 1
exact = sin(pi * z)
coef_f = (sin(pi * z) * (diff_cf * pi * pi * (1 - z * z) + reac_cf)
          + diff_cf * cos(pi * z) * 2 * pi * z)

# ----------------------------------- MAIN ------------------------------------
# Preliminary refinements
for i in range(n_cut_ref):
    lsetp1 = GridFunction(H1(mesh, order=1))
    InterpolateToP1(levelset, lsetp1)
    RefineAtLevelSet(lsetp1)
    mesh.Refine()

# Class to compute the mesh transformation needed for higher order accuracy
#  * order: order of the mesh deformation function
#  * threshold: barrier for maximum deformation (to ensure shape regularity)
lsetmeshadap = LevelSetMeshAdaptation(mesh, order=order, threshold=1000,
                                      discontinuous_qn=True)
deformation = lsetmeshadap.CalcDeformation(levelset)
lset_approx = lsetmeshadap.lset_p1

# Background FESpace
Vh = H1(mesh, order=order, dirichlet=[])

ci = CutInfo(mesh, lset_approx)
ba_IF = ci.GetElementsOfType(IF)
VhG = Restrict(Vh, ba_IF)

gfu = GridFunction(VhG)

# Coefficients / parameters:
n = Normalize(grad(lset_approx))
h = specialcf.mesh_size


# Tangential projection
def P(u):
    return u - (u * n) * n


u, v = VhG.TnT()

# Measure on surface
ds = dCut(lset_approx, IF, definedonelements=ba_IF, deformation=deformation)
# Measure on the bulk around the surface
dx = dx(definedonelements=ba_IF, deformation=deformation)

# Bilinear forms:
a = BilinearForm(VhG, symmetric=True)
a += (diff_cf * P(grad(u)) * P(grad(v)) + reac_cf * u * v) * ds
a += ((diff_cf / h + reac_cf * h) * (grad(u) * n) * (grad(v) * n)) * dx

# R.h.s. linear form
f = LinearForm(VhG)
f += coef_f * v * ds

a.Assemble()
f.Assemble()

gfu.vec.data = a.mat.Inverse(VhG.FreeDofs()) * f.vec

print("l2error : ", sqrt(Integrate((gfu - exact)**2 * ds, mesh=mesh)))
Draw(deformation, mesh, "deformation")
Draw(gfu, mesh, "u")

if len(sys.argv) < 2 or not (sys.argv[1] == "testmode"):
    input("Continue (press enter) to create a VTK-Output to tracefem3d.vtk")
    VTKOutput(ma=mesh, coefs=[deformation, lset_approx, gfu],
              names=["deformation", "P1-levelset", "u"],
              filename="tracefem", subdivision=2).Do()



# ------------------------------ LOAD LIBRARIES -------------------------------
import pytest
from netgen.geom2d import SplineGeometry
from ngsolve import *
from xfem import *
from xfem.lsetcurv import *

ngsglobals.msg_level = 2
SetNumThreads(4)

# -------------------------------- PARAMETERS ---------------------------------
lowerleft, upperright = (0.0,0.0), (2.2,0.41)   # Background domain corners
h_max = 0.07                                    # Mesh diameter
k = 5                                           # Order of velocity space
                                                
mapping = True                                  # Use isoparametric mapping 
nu = 1e-3                                       # Fluid viscosity
gamma_n = 40                                    # Nitsche parameter
gamma_gp = 0.01                                 # Ghost penalty parameter
pReg = 1e-8                                     # Pressure regularisation

inverse = ""                                    # Direct linear solver used
condense = True                                 # Use static condensation 
maxit_newt = 15                                 # Max. nr. of Newton iterations
tol_newt = 1e-10                                # l2 Newton residual tolerance
update_jacobi_tol = 0.15                        # Factor for Jacobean update
reuse_jacobi = True                             # Try to reuse Jacobean
print_newt = True                               # Newton convergence info


# -------------------------- CUTFEM HELPER FUNCTIONS --------------------------
def MarkCutElementsForCondensing(mesh, FES, ghost_facets):
    
    ghost_el = GetElementsWithNeighborFacets(mesh, ghost_facets)

    for el in mesh.Elements():
        if ghost_el[el.nr]:
            for dof_nr in FES.GetDofNrs(el):
                if FES.CouplingType(dof_nr) == COUPLING_TYPE.LOCAL_DOF:
                    FES.SetCouplingType(dof_nr, COUPLING_TYPE.INTERFACE_DOF)

    FES.FinalizeUpdate()


# -------------------------------- NEWTON LOOP --------------------------------
def CutFEM_QuasiNewton(a, alin, u, f, freedofs, maxit=100, maxerr=1e-11,
                       inverse="", dampfactor=1, jacobi_update_tol=0.1,
                       reuse = False, printing=True, **kwargs):

    res = u.vec.CreateVector()
    res_freedofs = u.vec.CreateVector()
    du = u.vec.CreateVector()
    numit = 0
    err,errLast = float("NaN"),float("NaN")
    Updated = "n/a "

    if reuse and not "inv_jacobian" in globals():
        global inv_jacobian
        JacobianAvailable = False
    if reuse and "inv_jacobian" in globals():
        JacobianAvailable = True
    else:
        JacobianAvailable = False

    projector = Projector(freedofs, True)

    if printing: 
        print("Numit\tUpd.J\t||res||_l2")
        print("--------------------------")

    for it in range(maxit):

        a.Assemble()
        res.data = a.mat * u.vec
        if f: res.data -= f
        res_freedofs.data = projector*res
        if it > 0:
            errLast = err
        err = Norm(res_freedofs)
        if printing: print ("{:}\t{:}\t{:1.4e}".format(numit,Updated,err))
        if err < maxerr: break
        elif not JacobianAvailable or err > errLast*jacobi_update_tol:
            UpdateJacobian = True
        else:
            UpdateJacobian = False

        numit += 1

        if UpdateJacobian:
            if JacobianAvailable:
                del inv_jacobian
            alin.Assemble()
            inv_jacobian = alin.mat.Inverse(freedofs, inverse=inverse)
            Updated = True
            JacobianAvailable = True
        else:
            Updated = False

        if alin.condense:
            res.data += alin.harmonic_extension_trans * res
            du.data = inv_jacobian * res
            du.data += alin.inner_solve * res
            du.data += alin.harmonic_extension * du
        else:
            du.data = inv_jacobian*res
            
        u.vec.data -= min(1,numit*dampfactor)*du

    else:
        print("Warning: Newton might not have converged!")
        return (-1,numit)

    if not reuse and JacobianAvailable: del inv_jacobian
    return (0,numit)


def test_st2d1_drag_lift():
    # ----------------------------------- DATA ------------------------------------
    def levelset_func(t):
        return 0.05 - sqrt((x - 0.2) * (x - 0.2) + (y - 0.2) * (y - 0.2))

    u_inflow = CoefficientFunction((4 * 0.3 * y * (0.41 - y) / (0.41**2), 0.0))

    # ------------------------------ BACKGROUND MESH ------------------------------
    geo = SplineGeometry()
    p1, p2, p3, p4, p5, p6 = [geo.AppendPoint(x,y) for x,y in [(0, 0), (0.7, 0), 
                                    (2.2, 0), (2.2, 0.41), (0.7,0.41), (0,0.41)] ]
    geo.Append (["line", p1, p2], leftdomain=1, rightdomain=0, bc="wall")
    geo.Append (["line", p2, p5], leftdomain=1, rightdomain=2)
    geo.Append (["line", p5, p6], leftdomain=1, rightdomain=0, bc="wall")
    geo.Append (["line", p6, p1], leftdomain=1, rightdomain=0, bc="inlet")
    geo.Append (["line", p2, p3], leftdomain=2, rightdomain=0, bc="wall")
    geo.Append (["line", p3, p4], leftdomain=2, rightdomain=0, bc="outlet")
    geo.Append (["line", p4, p5], leftdomain=2, rightdomain=0, bc="wall")

    geo.SetDomainMaxH(1, h_max/6)
    mesh = Mesh(geo.GenerateMesh(maxh=h_max))


    # --------------------------- FINITE ELEMENT SPACE ----------------------------
    V = VectorH1(mesh, order=k, dirichlet="wall|inlet")
    Q = H1(mesh, order=k-1)
    X = FESpace([V,Q], dgjumps=True)

    gfu = GridFunction(X)
    vel,pre = gfu.components
    gfu.vec[:] = 0.0


    # ---------------------------- LEVELSET & CUT-INFO ----------------------------
    # Levelset approximation
    if mapping:
        lset_meshadap = LevelSetMeshAdaptation(mesh, order=k, discontinuous_qn=True)
        deformation = lset_meshadap.CalcDeformation(levelset_func(0.0))
        mesh.SetDeformation(deformation)
        lsetp1 = lset_meshadap.lset_p1
    else:
        lsetp1 = GridFunction(H1(mesh,order=1))                     
        InterpolateToP1(levelset_func(0.0),lsetp1)

    # Integration dictionaries
    lset_neg = {"levelset": lsetp1, "domain_type": NEG, "subdivlvl": 0}
    lset_if = {"levelset": lsetp1, "domain_type": IF, "subdivlvl": 0}

    # Cut-info class
    ci_main = CutInfo(mesh, lsetp1)

    # ------------------------------ ELEMENT MARKERS ------------------------------
    els_hasneg = ci_main.GetElementsOfType(HASNEG)
    els_if = ci_main.GetElementsOfType(IF)
    facets_gp = GetFacetsWithNeighborTypes(mesh, a=els_hasneg, b=els_if,
                                           use_and=True, bnd_val_b=False)

    if condense:
        MarkCutElementsForCondensing(mesh, X, facets_gp)
    active_dofs = BitArray(X.ndof)
    active_dofs[:] = GetDofsOfElements(X, els_hasneg) & X.FreeDofs(coupling=condense)

    # ----------------------------- (BI)LINEAR FORMS ------------------------------
    (u, p), (v, q) = X.TnT()

    h = specialcf.mesh_size
    n_levelset = 1.0/Norm(grad(lsetp1)) * grad(lsetp1)

    stokes = nu * InnerProduct(Grad(u), Grad(v)) - p*div(v) - q*div(u)
    convect = InnerProduct(Grad(u) * vel, v)
    convect_lin = InnerProduct(Grad(u) * vel, v) + InnerProduct(Grad(vel) * u, v)

    nitsche  = -nu * InnerProduct(grad(u) * n_levelset,v)
    nitsche += -nu * InnerProduct(grad(v) * n_levelset,u)
    nitsche += nu * (gamma_n * k * k / h) * InnerProduct(u, v)
    nitsche += p * InnerProduct(v, n_levelset)
    nitsche += q * InnerProduct(u, n_levelset)

    ghost_penalty = gamma_gp * nu * (1 / h**2) * (u - u.Other()) * (v - v.Other())
    ghost_penalty += - gamma_gp * (1 / nu) * (p - p.Other()) * (q - q.Other())


    # -------------------------------- INTEGRATORS --------------------------------

    a = RestrictedBilinearForm(X, element_restriction=els_hasneg, 
                               facet_restriction=facets_gp, 
                               check_unused=False, flags={"symmetric": False})
    a += SymbolicBFI(lset_neg, form=stokes + convect, 
                     definedonelements=els_hasneg)
    a += SymbolicBFI(lset_if, form=nitsche, definedonelements=els_if)
    a += SymbolicFacetPatchBFI(form=ghost_penalty, skeleton=False, 
                               definedonelements=facets_gp)

    a_lin = RestrictedBilinearForm(X, element_restriction=els_hasneg, 
                                   facet_restriction=facets_gp, 
                                   check_unused=False, 
                                   flags={"condense":condense, 
                                          "symmetric": False})
    a_lin += SymbolicBFI(lset_neg, form=stokes - pReg * p * q + convect_lin,
                         definedonelements=els_hasneg)
    a_lin += SymbolicBFI(lset_if, form=nitsche, definedonelements=els_if)
    a_lin += SymbolicFacetPatchBFI(form=ghost_penalty, skeleton=False, 
                                   definedonelements=facets_gp)


    # ------------------------- SOLVE STATIONARY PROBLEM --------------------------
    with TaskManager():
        
        vel.Set(u_inflow, definedon=mesh.Boundaries("inlet"))
        CutFEM_QuasiNewton(a=a, alin=a_lin, u=gfu, f=None, freedofs=active_dofs, 
                           maxit=maxit_newt, maxerr=tol_newt, inverse=inverse,
                           jacobi_update_tol=update_jacobi_tol, reuse=reuse_jacobi,
                           printing=print_newt)


    # --------------------------- FUNTIONAL EVALUATION ----------------------------

    drag_x_test, drag_y_test = GridFunction(X), GridFunction(X)
    drag_x_test.components[0].Set(CoefficientFunction((1.0,0.0)))
    drag_y_test.components[0].Set(CoefficientFunction((0.0,1.0)))

    n = specialcf.normal(mesh.dim)
    a_test = BilinearForm(X, symmetric=False, check_unused=False)
    a_test += SymbolicBFI(lset_neg, form=convect, definedonelements=els_hasneg)
    a_test += SymbolicBFI(form=-InnerProduct(nu*Grad(u)*n - p*n,v), skeleton=True, 
                          definedon=mesh.Boundaries("inlet|wall"))
    with TaskManager():
        a_test.Assemble()

    res = gfu.vec.CreateVector()
    res.data = a_test.mat * gfu.vec
    Um = 2 * 0.3 / 3
    C_drag = -2.0 / (0.1 * Um**2) * InnerProduct(res, drag_x_test.vec)
    C_lift = -2.0 / (0.1 * Um**2) * InnerProduct(res, drag_y_test.vec)
    pdiff = pre(0.15, 0.2) - pre(0.25, 0.2)

    err_drag = abs(C_drag - 5.57953523384)
    err_lift = abs(C_lift - 0.010618948146)
    err_p = abs(pdiff - 0.11752016697)

    print("\n     C_drag      C_lift     pdiff")
    print("Val: {:10.8f}  {:10.8f}  {:10.8f}".format(C_drag, C_lift, pdiff))
    print("Err: {:1.2e}    {:1.2e}    {:1.2e}".format(err_drag, err_lift, err_p))

    assert err_drag < 2.5e-5
    assert err_lift < 2e-6
    assert err_p    < 2e-4

if __name__ == "__main__":
    test_st2d1_drag_lift()



"""
    Test: Integrate on an unfitted unit square
"""
# ------------------------------ LOAD LIBRARIES -------------------------------
from netgen.geom2d import SplineGeometry
from ngsolve import *
from xfem import *
from xfem.mlset import *

# -------------------------------- PARAMETERS ---------------------------------
h_max = 0.4
k = 4

function = x * (1 - x) * y * (1 - y)
value = 1 / 36

# ----------------------------------- DATA ------------------------------------


def level_sets():
    return [-y, x - 1, y - 1, -x]


nr_ls = len(level_sets())

# ------------------------------ BACKGROUND MESH ------------------------------
geo = SplineGeometry()
geo.AddRectangle((-0.2, -0.2), (1.2, 1.2),
                 bcs=("bottom", "right", "top", "left"))
mesh = Mesh(geo.GenerateMesh(maxh=h_max))

# ---------------------------- LEVELSET & CUT-INFO ----------------------------

level_sets_p1 = tuple(GridFunction(H1(mesh, order=1)) for i in range(nr_ls))
for i, lsetp1 in enumerate(level_sets_p1):
    InterpolateToP1(level_sets()[i], lsetp1)

square = DomainTypeArray([(NEG, NEG, NEG, NEG)])
Draw(square.Indicator(level_sets_p1), mesh, "mlsets")

lset_dom_inner = {"levelset": level_sets_p1, "domain_type": square}

# --------------------------------- INTEGRATE ---------------------------------
result = Integrate(levelset_domain=lset_dom_inner, mesh=mesh, cf=function, order=k)
print("Result: {}".format(result))
print("Error : {}".format(abs(result-value)))
assert abs(result-value) < 1e-12


"""
These examples show a number of pre-implemented 3d level set geometries,
computes the distance between the smooth level set and the (deformed) P1
approximation as well as the order of convergence thereof.

References:
-----------
* All concepts that are used here are explained in the jupyter-tuorials
  `basics` and `intlset`.
"""

# ------------------------------ LOAD LIBRARIES -------------------------------
from netgen.csg import CSGeometry
from ngsolve import *
from ngsolve.internal import *

from xfem import *
from xfem.lsetcurv import *
from xfem.utils import LevelsetExamples, BoundingBoxes

import sys

ngsglobals.msg_level = 2

# -------------------------------- PARAMETERS ---------------------------------
# Initial mesh size
maxh = 0.5
# Mesh deformation order
order = 2
# Refine cut elements
maxreflvl = 4

# ----------------------------------- MAIN ------------------------------------
# Main loop
for lsetgeom in ["cheese", "torus", "dziukelliott", "dziuk88", "sphere"]:
    print('Geometry: ', lsetgeom)

    geom = CSGeometry()
    geom.Add(BoundingBoxes[lsetgeom])
    mesh = Mesh(geom.GenerateMesh(maxh=maxh))
    levelset = LevelsetExamples[lsetgeom]

    lsetmeshadap = LevelSetMeshAdaptation(mesh, order=order, threshold=100,
                                          discontinuous_qn=True)

    Draw(levelset, mesh, "levelset")
    Draw(lsetmeshadap.deform, mesh, "deformation")
    Draw(lsetmeshadap.lset_p1, mesh, "levelset(P1)")

    distances = []
    for reflevel in range(maxreflvl):

        if (reflevel > 0):
            mesh.Refine()

        deformation = lsetmeshadap.CalcDeformation(levelset)
        distances.append(lsetmeshadap.CalcMaxDistance(levelset, deform=True))

        # Refine cut elements:
        RefineAtLevelSet(gf=lsetmeshadap.lset_p1)

        # Post-processing
        Redraw(blocking=True)
        eoc = [log(distances[i - 1] / distances[i]) / log(2)
               for i in range(1, len(distances))]
        print("distances = {}".format(distances))
        print("eoc = {}".format(eoc))

    if (hasattr(sys, 'ps1') or sys.flags.interactive
            or 'running_in_netgen' in dir()):
        input("press ENTER to go to next geometry (or finish)")



from ngsolve import *
from xfem import *
from netgen.geom2d import SplineGeometry

# --- Cut situation to be tested: ---
#
#   (0,1) 
#     \
#     |\
#     | \
#     |  \
#     |   \
#     +    \
#     |\    \
#     | \    \
#     |  \    \
#     |NEG\ POS\
#     +----+----+
#   (0,0)     (1,0)
#
# meas_2(POS) = 3/8, meas_2(NEG) = 1/8, meas_1(IF) = 1/2

# simple test: triangle, straight cut, integrand = 1
def test_cut_triangle():
    geom = SplineGeometry()
    pnts = [ (0,0), (1,0), (0,1) ]
    pnums = [geom.AppendPoint(*p) for p in pnts]
    lines = [(0,1,1,1,0), (1,2,1,1,0), (2,0,1,1,0)]
    for p1,p2,bc,left,right in lines:
        geom.Append( ["line", pnums[p1], pnums[p2]], bc=bc, leftdomain=left, rightdomain=right)
    mesh = Mesh (geom.GenerateMesh(maxh=1)) # <- will have 4 elements

    levelset = x + y - 0.25
    lsetp1 = GridFunction(H1(mesh,order=1))
    InterpolateToP1(levelset,lsetp1)

    lset_neg = { "levelset" : lsetp1, "domain_type" : NEG, "subdivlvl" : 0}
    lset_pos = { "levelset" : lsetp1, "domain_type" : POS, "subdivlvl" : 0}

    for order in range(16):
        measure_neg = Integrate(levelset_domain = lset_neg,cf=CoefficientFunction(1.0), mesh = mesh, order=order)
        measure_pos = Integrate(levelset_domain = lset_pos,cf=CoefficientFunction(1.0), mesh = mesh, order=order)
        assert abs(measure_neg-1.0/32.0) < 5e-16*(order+1)*(order+1)
        assert abs(measure_pos-1.0/2.0+1.0/32.0) < 5e-16*(order+1)*(order+1)
        assert abs(measure_neg+measure_pos-1.0/2.0) < 5e-16*(order+1)*(order+1)




import pytest
from ngsolve import *
from xfem import *
from ngsolve.meshes import *
from netgen.geom2d import SplineGeometry
from netgen.csg import CSGeometry, OrthoBrick, Pnt
from math import pi
from xfem.lsetcurv import *

def test_calc_linearized():
    square = SplineGeometry()
    square.AddRectangle([-1.5,-1.5],[1.5,1.5],bc=1)
    mesh = Mesh (square.GenerateMesh(maxh=0.8))
    
    levelset = (sqrt(sqrt(x*x*x*x+y*y*y*y)) - 1.0)

    lsetp1 = GridFunction(H1(mesh))
    lsetp1.Set(levelset)
    lset_neg = { "levelset" : lsetp1, "domain_type" : NEG, "subdivlvl" : 0}
    
    ci = CutInfo(mesh,lsetp1)
    hasneg = ci.GetElementsOfType(HASNEG)

    Vh = H1(mesh, order = 2, dirichlet=[])    
    active_dofs = GetDofsOfElements(Vh,hasneg)
    Vh = Compress(Vh,active_dofs)
    
    u,v = Vh.TnT()

    gfu = GridFunction(Vh)
    gfu.Set(sin(x))
    
    a1 = BilinearForm(Vh)
    a1 += SymbolicBFI(levelset_domain = lset_neg, form = u*u * v)
    a1.AssembleLinearization(gfu.vec)

    a2 = BilinearForm(Vh)
    a2 += SymbolicBFI(levelset_domain = lset_neg, form = 2*gfu*u * v)
    a2.Assemble()

    a3 = BilinearForm(Vh)
    a3 += SymbolicBFI(levelset_domain = lset_neg, form = gfu*u * v)
    
    w1 = gfu.vec.CreateVector()
    w2 = gfu.vec.CreateVector()

    w1.data = a1.mat * gfu.vec
    w2.data = a2.mat * gfu.vec
    
    w1.data -= w2
    diff = Norm(w1)
    print("diff : ",diff)
    assert diff < 1e-12

    a1.Apply(gfu.vec,w1)
    a3.Apply(gfu.vec,w2)
    
    w1.data -= w2
    diff = Norm(w1)
    print("diff : ",diff)
    assert diff < 1e-12
    
if __name__ == "__main__":
    test_calc_linearized()
    



"""
In this example we solve a surface stokes problem with a
consistent penalty method following [1]

Used features:
--------------
* Higher order geometry approximation, cf. jupyter tutorial `lsetint`

* Restricted finite element space to condense the system to active dofs,
  cf. jupyter tutorial `basics`

* Visualization: The visualization of the solution is most convenient
  with paraview and the generated vtk file.

Literature:
-----------
[1] T. Jankuhn, A. Reusken, Higher order Trace Finite Element Methods for the Surface Stokes Equation.
    arXiv preprint arXiv:1909.08327, 2019.

"""

# ------------------------------ LOAD LIBRARIES -------------------------------
import sys

from netgen.csg import CSGeometry, OrthoBrick, Pnt
from ngsolve import *
from ngsolve.internal import *
from xfem import *
from xfem.lsetcurv import *
from math import pi
import time
# -------------------------------- PARAMETERS ---------------------------------
# Interpolation for the rate-of-strain tensor when computing the r.h.s.
# Introduces an interpolation error, but significantly improves performance
interpol = True
# Mesh diameter
maxh = 0.6
# Geometry
geom = "ellpsoid"
# Polynomial order of FE space
order = 2

# Problem parameters
reac_cf = 1
diff_cf = 1
# Ellipsoid parameter
c = 1.4
# Geometry
with TaskManager():
    cube = CSGeometry()
    if geom == "ellpsoid":
        phi = Norm(CoefficientFunction((x / c, y, z))) - 1

        cube.Add(OrthoBrick(Pnt(-2, -2, -2), Pnt(2, 2, 2)))
        mesh = Mesh(cube.GenerateMesh(maxh=maxh, quad_dominated=False))
    elif geom == "decocube":
        phi = (x ** 2 + y ** 2 - 4) ** 2 + (y ** 2 - 1) ** 2 + (y ** 2 + z ** 2 - 4) ** 2 + (x ** 2 - 1) ** 2 + (
                    x ** 2 + z ** 2 - 4) ** 2 + (z ** 2 - 1) ** 2 - 13
        cube.Add(OrthoBrick(Pnt(-3, -3, -3), Pnt(3, 3, 3)))
        mesh = Mesh(cube.GenerateMesh(maxh=maxh, quad_dominated=False))

    # Preliminary refinements. As assembling the r.h.s. is very expensive without interpolation, keep it small in that case.
    # When interpolating, getting good results requires more refinements with complex geometries, e.g. the decocube
    n_cut_ref = 2

    for i in range(n_cut_ref):
        lsetp1 = GridFunction(H1(mesh, order=2))
        InterpolateToP1(phi, lsetp1)
        RefineAtLevelSet(lsetp1)
        mesh.Refine()
    # Class to compute the mesh transformation needed for higher order accuracy
    #  * order: order of the mesh deformation function
    #  * threshold: barrier for maximum deformation (to ensure shape regularity)

    lsetmeshadap = LevelSetMeshAdaptation(mesh, order=order, threshold=1000, discontinuous_qn=True)
    deformation = lsetmeshadap.CalcDeformation(phi)
    lset_approx = lsetmeshadap.lset_p1

    mesh.SetDeformation(deformation)

# Background FESpaces
    Vh = VectorH1(mesh, order=order, dirichlet=[])
    Qh = H1(mesh, order=order - 1, dirichlet=[])
    ci = CutInfo(mesh, lset_approx)
    ba_IF = ci.GetElementsOfType(IF)
    VhG = Restrict(Vh, ba_IF)
    L2G = Restrict(Qh, ba_IF)

    X = VhG * L2G
    gfu = GridFunction(X)


    # Helper Functions for tangential projection and gradients of the exact solution
    def coef_grad(u):
        dirs = {1: [x], 2: [x, y], 3: [x, y, z]}
        if u.dim == 1:
            return CF(tuple([u.Diff(r) for r in dirs[mesh.dim]]))
        else:
            return CF(tuple([u[i].Diff(r) for i in range(u.dim) for r in dirs[mesh.dim]]), dims=(u.dim, mesh.dim))


    def grad(u):
        if type(u) in [ProxyFunction, GridFunction]:
            return ngsolve.grad(u)
        else:
            return coef_grad(u)


    def Pmats(n):
        return Id(3) - OuterProduct(n, n)


    # Coefficients / parameters:

    n = Normalize(grad(lset_approx))
    ntilde = Interpolate(grad(phi), VhG)

    h = specialcf.mesh_size

    eta = 100.0 / (h * h)
    rhou = 1.0 / h
    rhop = h

    # Measure on surface
    ds = dCut(lset_approx, IF, definedonelements=ba_IF, deformation=deformation)
    # Measure on the bulk around the surface
    dx = dx(definedonelements=ba_IF, deformation=deformation)

    # Exact tangential projection
    Ps = Pmats(Normalize(grad(phi)))
    # define solution and right-hand side

    if geom == "ellpsoid":
        functions = {
            "extvsol1": ((-y - z) * x + y * y + z * z),
            "extvsol2": ((-x - z) * y + x * x + z * z),
            "extvsol3": ((-x - y) * z + x * x + y * y),
            "rhs1": -((y + z) * x - y * y - z * z) * (x * x + y * y + z * z + 1) / (x * x + y * y + z * z),
            "rhs2": ((-x - z) * y + x * x + z * z) * (x * x + y * y + z * z + 1) / (x * x + y * y + z * z),
            "rhs3": ((-x - y) * z + x * x + y * y) * (x * x + y * y + z * z + 1) / (x * x + y * y + z * z),
        }
        pSol = (x * y ** 3 + z * (x ** 2 + y ** 2 + z ** 2) ** (3 / 2)) / ((x ** 2 + y ** 2 + z ** 2) ** 2)
        uSol = Ps* CoefficientFunction((functions["extvsol1"], functions["extvsol2"], functions["extvsol3"]))

    elif geom == "decocube":
        uSol = Ps*CoefficientFunction((-z ** 2, y, x))
        pSol = CoefficientFunction(
            x * y ** 3 + z - 1 / Integrate({"levelset": lset_approx, "domain_type": IF}, cf=CoefficientFunction(1),
                                        mesh=mesh) * Integrate({"levelset": lset_approx, "domain_type": IF},
                                                                cf=x * y ** 3 + z, mesh=mesh))



    u, p = X.TrialFunction()
    v, q = X.TestFunction()


    # Helper Functions for rate-of-strain-tensor
    def eps(u):
        return Ps * Sym(grad(u)) * Ps


    def divG(u):
        if u.dim == 3:
            return Trace(grad(u) * Ps)
        if u.dim == 9:
            if u.dims[0] == 3:
                N = 3
                divGi = [divG(u[i, :]) for i in range(N)]
                return CF(tuple([divGi[i] for i in range(N)]))
            else:
                N = 3
                r = Grad(u) * Ps
                return CF(tuple([Trace(r[N*i:N*(i+1),0:N]) for i in range(N)]))

                #divGi = [divG(u[i, :]) for i in range(N)]
                #return CF(tuple([divGi[i] for i in range(N)]))
    # set the r.h.s. either with interpolation or exactly
    def rhsfromsol(uSol, pSol, interpol=True):
        if interpol:
            tmp = GridFunction(Restrict(L2(mesh, order=order+2, dim=9), ba_IF))
            tmp.Set(eps(uSol) ,definedonelements=ba_IF)
            tmp2 = GridFunction(Restrict(VectorL2(mesh, order=order+1), ba_IF))
            tmp2.Set(Ps*-divG(tmp)+uSol+Ps*grad(pSol), definedonelements=ba_IF)
            return tmp2
        else:
            
            return Ps * (-divG(eps(uSol).Compile())) + uSol +Ps* grad(pSol).Compile()

    # Weingarten mappings


    ninter = GridFunction(VhG)
    ninter.Set(n, definedonelements=ba_IF)
    weing = grad(ninter)
    weingex = grad(Normalize(grad(phi)))

    rhs1 = rhsfromsol(uSol, pSol, interpol)
    rhs2 = Trace(Ps * grad(uSol))
    Pmat = Pmats(n)


    def E(u):
        return Pmat * Sym(grad(u)) * Pmat# - (u * n) * weing


    a = BilinearForm(X, symmetric=True)
    a += InnerProduct(E(u), E(v)) * ds
    a += (Pmat * u) * (Pmat * v) * ds
    a += (eta * ((u * ntilde) * (v * ntilde))) * ds
    a += rhou * ((grad(u) * n) * (grad(v) * n)) * dx
    a += InnerProduct(u, Pmat * grad(q)) * ds
    a += InnerProduct(v, Pmat * grad(p)) * ds
    a += -rhop * InnerProduct(n * grad(p), n * grad(q)) * dx

    a.Assemble()

    # R.h.s. linear form

    f = LinearForm(X)
    f += rhs1 * v * ds
    f += -rhs2 * q * ds
    start=time.time()
    f.Assemble()
    
    gfu.vec.data = a.mat.Inverse(freedofs=X.FreeDofs()) * f.vec

    print("l2error : ", sqrt(Integrate((gfu.components[0] - uSol) ** 2 * ds, mesh=mesh)))
uerr = (gfu.components[0] - uSol)

Draw(deformation, mesh, "deformation")
Draw(gfu.components[0], mesh, "u")

if len(sys.argv) < 2 or not (sys.argv[1] == "testmode"):
    input("Continue (press enter) to create a VTK-Output to stokestracefem3d.vtk")

    with TaskManager():
        vtk = VTKOutput(ma=mesh,
                        coefs=[lset_approx, gfu.components[0], uSol, uerr],
                        names=["P1-levelset", "uh", "uex", "uerr"],
                        filename="stokestracefem3d", subdivision=2)

        vtk.Do()



"""
integration on levelset domains
"""

# ------------------------------ LOAD LIBRARIES -------------------------------
from netgen.csg import CSGeometry, OrthoBrick, Pnt
from ngsolve import *
from xfem import *


# ----------------------------------- MAIN ------------------------------------
cube = OrthoBrick(Pnt(0, 0, 0), Pnt(1, 1, 1)).bc(1)
geom = CSGeometry()
geom.Add(cube)
mesh = Mesh(geom.GenerateMesh(maxh=1, quad_dominated=True))


def binary_pow(x, a):
    if a == 0:
        return 0 * x + 1
    elif a == 1:
        return x
    else:
        print("Invalid argument a")


# levelset coefficients
c = [[[1, -2], [-2, 0]], [[-2, 0], [0, 0]]]

# levelset = 1-2*x-2*y-2*z #(sqrt(x*x+y*y+z*z)-0.5)
levelset = sum([c[alpha][beta][gamma] * binary_pow(x, alpha)
                * binary_pow(y, beta) * binary_pow(z, gamma)
                for alpha in [0, 1] for beta in [0, 1] for gamma in [0, 1]])
referencevals = {POS: 1. / 48, NEG: 47. / 48, IF: sqrt(3) / 8}

V = H1(mesh, order=1)
lset_approx = GridFunction(V)
InterpolateToP1(levelset, lset_approx)

f = CoefficientFunction(1.0)

domains = [NEG, POS, IF]

errors = dict()
eoc = dict()

for key in domains:
    errors[key] = []
    eoc[key] = []

for order in range(8):
    for key in domains:
        lset_dom = {"levelset": lset_approx, "domain_type": key}
        integral = Integrate(levelset_domain=lset_dom,
                             cf=f, mesh=mesh, order=order)
        print("Integral on Domain ", key, " : ", integral)
        errors[key].append(abs(integral - referencevals[key]))

Draw(levelset, mesh, "lset")
print("L2 Errors:", errors)



from netgen.geom2d import SplineGeometry
from netgen.meshing import MeshingParameters
from ngsolve import *
from xfem import *
# For LevelSetAdaptationMachinery
from xfem.lsetcurv import *


r2 = 3/4 # outer radius
r1 = 1/4 # inner radius
rc = (r1+r2) / 2.0
rr = (r2-r1) / 2.0
r = sqrt(x*x+y*y)
levelset = IfPos(r-rc, r-rc-rr,rc-r-rr)

# Geometry 
square = SplineGeometry()
square.AddRectangle([-1,-1],[1,1],bc=1)

def meshonlvl(lvl):
    ngmesh = square.GenerateMesh(maxh=0.9, quad_dominated=False)
    mesh = Mesh (ngmesh)
    for i in range(lvl):
        mesh.Refine()
    return mesh


#trig version:
ngmesh = square.GenerateMesh(maxh=0.9, quad_dominated=False)
mesh = Mesh (ngmesh)
p1prolong = P1Prolongation(mesh)

prolongNeg = P1Prolongation(mesh)
prolongPos = P1Prolongation(mesh)



order = 1
nref = 2

Vh = H1(mesh, order = order, dirichlet=[], dgjumps = True)
lsetp1 = GridFunction(Vh)
InterpolateToP1(levelset,lsetp1)

def GetActiveDof(mesh,HASPOSNEG):
    Vh = H1(mesh, order = order, dirichlet=[], dgjumps = True)
    lsetp1 = GridFunction(Vh)
    InterpolateToP1(levelset,lsetp1)
    ci = CutInfo(mesh,lsetp1)
    return GetDofsOfElements(Vh,ci.GetElementsOfType(HASPOSNEG))

VhNeg = Compress( Vh, active_dofs = GetActiveDof(mesh, HASNEG) )
VhPos = Compress( Vh, active_dofs = GetActiveDof(mesh, HASPOS) )

CutVh = FESpace( [VhNeg, VhPos] )

CutProl = CompoundProlongation( CutVh )

CutProl.AddProlongation(prolongNeg)
CutProl.AddProlongation(prolongPos)

Vhc = Compress(Vh,active_dofs=GetActiveDof(mesh,HASNEG))
#p1prolong.Update(Vhc)
prolongNeg.Update(VhNeg)
prolongPos.Update(VhPos)
gfu0 = GridFunction( CutVh )

gfu0.components[0].Set( CoefficientFunction(x*x*y) )
gfu0.components[1].Set( CoefficientFunction(x*x*y+0.1) )
tmpvec = gfu0.vec.CreateVector()
for i in range(len(tmpvec)):
    tmpvec[i] = gfu0.vec[i]

cutgfu = GridFunction( CutVh )
cutgfu.vec[:] = 0.0

gfu = GridFunction(Vhc)
print("ndof",Vhc.ndof)
print( "dim vneg",CutVh.components[0].ndof )
print( "dim vpos",CutVh.components[1].ndof )
dofArr = [ 0 for i in range(nref+1) ]
dofArr[0] = CutVh.ndof
gf = []
gf.append(GridFunction(Compress(H1(mesh,order=1),active_dofs=GetActiveDof(mesh,HASNEG))))
for i in range(nref):
    mesh.Refine()
    Vh.Update()
    lsetp1.Update()
    InterpolateToP1(levelset,lsetp1)
    ci = CutInfo(mesh,lsetp1)
    Vhc.SetActiveDofs(GetActiveDof(mesh,HASNEG))
    Vhc.Update()

    CutVh.components[0].SetActiveDofs( GetActiveDof(mesh,HASNEG) )
    CutVh.components[1].SetActiveDofs( GetActiveDof(mesh,HASPOS) )
    CutVh.Update()
    CutProl.Update(CutVh)

    dofArr[i+1] = CutVh.ndof

    print( "dim vneg",CutVh.components[0].ndof )
    print( "dim vpos",CutVh.components[1].ndof )
    gfu.Update()
    cutgfu.Update()
    #p1prolong.Update(Vhc)

    
    nmesh = meshonlvl(i)
    gf.append(GridFunction(Compress(H1(nmesh,order=1),active_dofs=GetActiveDof(nmesh,HASNEG))))

gfdraw = GridFunction( CutVh )    

v = gfdraw.vec.CreateVector()    
w = gfdraw.vec.CreateVector()

Pv = gfdraw.vec.CreateVector()    
Rw = gfdraw.vec.CreateVector()

for i in range(len(v)):
    v[i] = i
for i in range(len(w)):
    w[i] = 1.0/(i+1)

Rw.data = w    
CutProl.Restrict(2,Rw)
CutProl.Restrict(1,Rw)

#v[dofArr[0]: ] = 0.0
Pv.data = v
CutProl.Prolongate(1,Pv)
CutProl.Prolongate(2,Pv)

# print(v)
# print(Pv)
# print(w)
# print(Rw)

print(InnerProduct(Pv,w))
print(InnerProduct(v,Rw))

input('w')

    
v = gfu.vec.CreateVector()    
w = gfu.vec.CreateVector()


prolvec = cutgfu.vec.CreateVector()
for i in range( len(tmpvec) ):
    prolvec[i] = tmpvec[i]


print('before prol')
for i in range(nref):
    CutProl.Prolongate(i+1,prolvec)


offset = [ 0, len(gfdraw.components[0].vec) ]
lens   = [ len(gfdraw.components[i].vec) for i in [0,1] ]

for i in [0,1]:
    for k in range(offset[i], offset[i]+lens[i] ):
        gfdraw.components[i].vec[k-offset[i]] = prolvec[k]

u_coef = IfPos(lsetp1, gfdraw.components[1], gfdraw.components[0])
Draw(u_coef,mesh,"u")


input('w')

for i in range(len(v)):
    v[i] = i
w.data = v
# Pv = gfu.vec.CreateVector()    
# Rw = gfu.vec.CreateVector()


for i in range(len(gf[0].vec)):
    gf[0].vec[i] = i


for i in range(nref+1):
    if i > 0:
       p1prolong.Prolongate(i,w) 
    #print(w)
    nmesh = meshonlvl(i)
    gfvis = GridFunction(Compress(H1(nmesh,order=1),active_dofs=GetActiveDof(nmesh,HASNEG)))
    for i in range(len(gfvis.vec)):
        gfvis.vec[i] = w[i]
    Draw(gfvis,nmesh,"gf")
    input(i)
    
    
# v = gfu.vec.CreateVector()    
# w = gfu.vec.CreateVector()

# Pv = gfu.vec.CreateVector()    
# Rw = gfu.vec.CreateVector()

# for i in range(len(v)):
#     v[i] = i
# for i in range(len(w)):
#     w[i] = 1.0/(i+1)

# Rw.data = w    
# p1prolong.Restrict(2,Rw)
# p1prolong.Restrict(1,Rw)

# Pv.data = v
# p1prolong.Prolongate(1,Pv)
# p1prolong.Prolongate(2,Pv)

# print(InnerProduct(Pv,w))
# print(InnerProduct(v,Rw))
    
# # gfu.vec[:] = 1.0 

# # input("")
# print(gfu.vec)

# p1prolong.Restrict(2,gfu.vec)
# print(gfu.vec)
# input("")
# p1prolong.Restrict(1,gfu.vec)

# print(gfu.vec)
# input("")




from ngsolve import *
from time import sleep
from netgen.geom2d import unit_square
from netgen.geom2d import SplineGeometry
from netgen.meshing import MeshingParameters

from ngsolve.internal import *
from xfem import *
from numpy import pi

from xfem.lset_spacetime import *

# geometry
periodic = SplineGeometry()
pnts = [ (0,0), (2,0), (2,2), (0,2) ]
pnums = [periodic.AppendPoint(*p) for p in pnts]

uright = periodic.Append ( ["line", pnums[0], pnums[1]], bc="periodic")
periodic.Append ( ["line", pnums[3], pnums[2]], leftdomain=0, rightdomain=1, copy=uright, bc="periodic")

lright = periodic.Append ( ["line", pnums[1], pnums[2]], bc="periodic")
periodic.Append ( ["line", pnums[0], pnums[3]], leftdomain=0, rightdomain=1, copy=lright, bc="periodic")

maxh = 0.1
mesh = Mesh(periodic.GenerateMesh(maxh=maxh))

n_ref = 0
for i in range(n_ref):
    mesh.Refine()

h = specialcf.mesh_size

# FE Spaces
k_t = 2
k_s = 2
fes1 = Periodic(H1(mesh,order=k_s,dirichlet=[]))
fes_lset_slice = H1(mesh, order=1, dirichlet=[])
fes_dfm_slice = H1(mesh, order=k_s, dim=mesh.dim)


lset_order_time = 2
time_order = 4
tfe = ScalarTimeFE(k_t) 

st_fes = SpaceTimeFESpace(fes1,tfe)
VhG_ic = FESpace([fes1,fes1])
VhG = FESpace([st_fes,st_fes])
#visoptions.deformation = 1

tend = 1
delta_t = 1/16
tnew = 0

told = Parameter(0)
tref = ReferenceTimeVariable()
t = told + delta_t*tref

lset_adap_st = LevelSetMeshAdaptation_Spacetime(mesh, order_space = k_s, order_time = lset_order_time,
                                         threshold=0.5, discontinuous_qn=True)

# Problem data

problems =  {"planar": ( CoefficientFunction(0.25*t), CoefficientFunction(0.25),
                         CoefficientFunction(1),CoefficientFunction(0),CoefficientFunction(0)),
             "nonlinear": ( CoefficientFunction((0.25/pi)*sin(2*pi*t)),CoefficientFunction(0.5*cos(2*pi*t)),
                            CoefficientFunction(7/8 + (1/4)*y*y*(2-y)*(2-y)),CoefficientFunction( y*(1-y)*(2-y) ),CoefficientFunction( (1-y)*(2-y) - y*(2-y) - y*(1-y))) }
          
rho,d_rho,q_coeff,d_q_coeff,dd_q_coeff = problems["planar"]

#rho = CoefficientFunction(0.0)
#d_rho = CoefficientFunction(0.0)

alpha_neg = 1.0
alpha_pos = 2.0
alpha = [alpha_neg,alpha_pos]
beta_neg = 1.5
beta_pos = 1.0
beta = [beta_neg,beta_pos]

a_const = 1.02728
b_const = 6.34294
k_const = 1

stab = 10*(alpha[0]+alpha[1])*(k_s+1)*(k_s+1)/h

r11 = x - q_coeff - rho

w = CoefficientFunction((d_rho,0)) # convection
D_const = 2/3
levelset= CoefficientFunction(sqrt(r11*r11) - 0.5*D_const)


f_neg = ( k_const*pi*cos(k_const*pi*t)*(a_const*r11 + b_const*r11*r11*r11 )
         - alpha[0]*sin(k_const*pi*t)*6*b_const*r11*(1+d_q_coeff*d_q_coeff)
         + alpha[0] * sin(k_const*pi*t)*(a_const+3*b_const*r11*r11) * dd_q_coeff )

f_pos = ( k_const*pi*cos(k_const*pi*t)*sin(pi*r11)
         - alpha[1]*sin(k_const*pi*t)*(-pi*pi*sin(pi*r11))*(1+d_q_coeff*d_q_coeff)
         + alpha[1]*sin(k_const*pi*t)*pi*cos(pi*r11) * dd_q_coeff )

coeff_f = [f_neg,f_pos]

# for monitoring the error
exact = [sin(k_const*pi*t)*(a_const*r11+b_const*r11*r11*r11),
         sin(k_const*pi*t)*sin(pi*r11)]



u0 = GridFunction(VhG)
u0_ic = GridFunction(VhG_ic)
#Draw(u0_ic,mesh,"u")

# Definition of Trial/Test-functions
u = VhG.TrialFunction()
v = VhG.TestFunction()

gradu = [grad(ui) for ui in u]
gradv = [grad(vi) for vi in v]


lset_top = GridFunction(fes_lset_slice)
lset_bottom = GridFunction(fes_lset_slice)
dfm_top = GridFunction(fes_dfm_slice)

t_old = 0
coef_u0 = [CoefficientFunction(0),CoefficientFunction(0)]
u0_ic.components[0].Set(coef_u0[0]) # set initial condition
u0_ic.components[1].Set(coef_u0[1]) # set initial condition

m_reg = BilinearForm(VhG,symmetric=False)
epsilon = 1e-8
#epsilon = 0
m_reg.Assemble()
for ti in st_fes.TimeFE_nodes():
    m_tmp = BilinearForm(VhG,symmetric=False)
    m_tmp += SymbolicBFI(form = u[0]*v[0] + u[1]*v[1]) 
    st_fes.SetTime(ti)
    m_tmp.Assemble()
    m1 = m_tmp.mat.CreateMatrix()
    m_reg.mat.AsVector().data +=  m1.AsVector()
    
st_fes.SetOverrideTime(False)


while tend - t_old > delta_t/2:
    
    dfm = lset_adap_st.CalcDeformation(levelset,told,t_old,delta_t,calc_kappa = True) 
    dfm_top.vec[:] = dfm.vec[lset_order_time*lset_adap_st.ndof_node : (lset_order_time+1)*lset_adap_st.ndof_node]
    lset_p1 = lset_adap_st.lset_p1  

    active_dofs = BitArray(VhG.FreeDofs())
   
    hasneg_spacetime = lset_adap_st.hasneg_spacetime
    hasneg_spacetime |= lset_adap_st.hasif_spacetime
    haspos_spacetime = lset_adap_st.haspos_spacetime
    haspos_spacetime |= lset_adap_st.hasif_spacetime
    
    active_dofs &= CompoundBitArray([GetDofsOfElements(st_fes,hasneg_spacetime),GetDofsOfElements(st_fes,haspos_spacetime)])    

    lset_neg = { "levelset" : lset_p1, "domain_type" : NEG, "subdivlvl" : 0}
    lset_pos = { "levelset" : lset_p1, "domain_type" : POS, "subdivlvl" : 0}
    lset_if = { "levelset" : lset_p1, "domain_type" : IF, "subdivlvl" : 0}
    # Put this into LevelSetMeshAdaptation_Spacetime later on
    lset_bottom.vec[:] = lset_p1.vec[0:fes_lset_slice.ndof]
    lset_neg_bottom = { "levelset" : lset_bottom, "domain_type" : NEG, "subdivlvl" : 0}
    lset_pos_bottom = { "levelset" : lset_bottom, "domain_type" : POS, "subdivlvl" : 0}
    lset_top.vec[:] = lset_p1.vec[lset_order_time*fes_lset_slice.ndof : (lset_order_time+1)*fes_lset_slice.ndof]
    lset_neg_top = { "levelset" : lset_top, "domain_type" : NEG, "subdivlvl" : 0}
    lset_pos_top = { "levelset" : lset_top, "domain_type" : POS, "subdivlvl" : 0}
    
    n_lset_p1 = 1.0/grad(lset_adap_st.lset_p1).Norm() * grad(lset_adap_st.lset_p1)
    
    kappas = {"interp" : (lset_adap_st.kappa,1-lset_adap_st.kappa), 
              "st_CutInfo" : (CutRatioGF(lset_adap_st.ci),1-CutRatioGF(lset_adap_st.ci)) }
    kappa = kappas["interp"]
    
#    time_quad = lset_adap_st.v_ho_st.TimeFE_nodes().NumPy()
#    times = [t_old + delta_t * xi for xi in time_quad]
#    for tau in times:  
#        lset_adap_st.v_kappa.SetTime(tau)
#        Draw(kappa[0],mesh,"kappa")
#        input("Kappa at t = {0}".format(tau))
#    lset_adap_st.v_kappa.SetOverrideTime(False)
         
    average_flux_u = sum( [-kappa[i]*alpha[i]*gradu[i]*n_lset_p1 for i in [0,1]] )
    average_flux_v = sum( [-kappa[i]*alpha[i]*gradv[i]*n_lset_p1 for i in [0,1]] )
    betajump_u = beta[0]*u[0]-beta[1]*u[1]
    betajump_v = beta[0]*v[0]-beta[1]*v[1]
    
     
    a = BilinearForm(VhG,symmetric=False)

    a += SymbolicBFI(levelset_domain = lset_neg, form =  -u[0]*dt(v[0],0)*beta[0]
                                                      + delta_t*alpha[0]*beta[0]*gradu[0]*gradv[0]
                                                      - delta_t*beta[0]*u[0]*gradv[0]*w, time_order=time_order)    
                                                      
    a += SymbolicBFI(levelset_domain = lset_pos, form = -u[1]*dt(v[1],1)*beta[1]  
                                                      + delta_t*alpha[1]*beta[1]*gradu[1]*gradv[1]
                                                      - delta_t*beta[1]*u[1]*gradv[1]*w, time_order=time_order)
 
    
    a += SymbolicBFI(levelset_domain = lset_neg_top, form = beta[0]*fix_t(u[0],1,0)*fix_t(v[0],1,0) )
    a += SymbolicBFI(levelset_domain = lset_pos_top, form = beta[1]*fix_t(u[1],1,1)*fix_t(v[1],1,1) )
    
    

    a += SymbolicBFI(levelset_domain = lset_if, form = delta_t*( average_flux_u*betajump_v   
                                                              + betajump_u*average_flux_v 
                                                            + stab*betajump_u*betajump_v), time_order=time_order )

                                                               
                
    f = LinearForm(VhG)
    f += SymbolicLFI(levelset_domain = lset_neg, form = delta_t*beta[0]*coeff_f[0]*v[0], time_order=time_order)
    f += SymbolicLFI(levelset_domain = lset_pos, form = delta_t*beta[1]*coeff_f[1]*v[1], time_order=time_order)
    f += SymbolicLFI(levelset_domain = lset_neg_bottom,form = beta[0]*u0_ic.components[0]*fix_t(v[0],0,0) )
    f += SymbolicLFI(levelset_domain = lset_pos_bottom,form = beta[1]*u0_ic.components[1]*fix_t(v[1],0,1) )
    
    
    mesh.SetDeformation(dfm)
    a.Assemble()
    a.mat.AsVector().data += epsilon * m_reg.mat.AsVector()
    f.Assemble()
    mesh.UnsetDeformation()
    u0.vec.data = a.mat.Inverse(active_dofs,"") * f.vec
       
    # exploiting the nodal property of the time fe:
    for i in range(2): 
        u0_ic.components[i].vec[:] = u0.components[i].vec[k_t*fes1.ndof : (k_t+1)*fes1.ndof]
   
    t_old = t_old + delta_t
    told.Set(t_old)
    
    mesh.SetDeformation(dfm_top)
    err_sqr_coefs = [ (u0_ic.components[i] - exact[i])*(u0_ic.components[i] - exact[i]) for i in [0,1] ]
    l2error = sqrt( Integrate( levelset_domain=lset_neg_top, cf=err_sqr_coefs[0], mesh=mesh,
                           order=2*k_s, heapsize=1000000)
                + Integrate(levelset_domain=lset_pos_top, cf=err_sqr_coefs[1], mesh=mesh,
                            order=2*k_s, heapsize=1000000)) 
    mesh.UnsetDeformation()
    print("t = {0}, l2error = {1}".format(t_old,l2error))
    
    Draw(IfPos(lset_top,u0_ic.components[1],u0_ic.components[0]),mesh,"u")
#    Draw(IfPos(lset_top,sqrt(err_sqr_coefs[1]),sqrt(err_sqr_coefs[0])),mesh,"error")
#    Draw(IfPos(lset_top,exact[1],exact[0]),mesh,"exact")
    #input("Continue")
    #Redraw(blocking=True)
       


"""
integration on levelset domains
"""

# ------------------------------ LOAD LIBRARIES -------------------------------
from ngsolve import *
from xfem import *
from sympy import *


# ----------------------------------- MAIN ------------------------------------
def get_levelset(lsetvals):
    lset = lsetvals[0] + (lsetvals[1] - lsetvals[0]) * x
    lset += (lsetvals[3] - lsetvals[0]) * y
    lset += (lsetvals[2] - lsetvals[1] - lsetvals[3] + lsetvals[0]) * x * y
    return lset


def get_referencevals(lsetvals, f, IF_INT=False):
    d = lsetvals[0]
    c = lsetvals[2] - lsetvals[1] - lsetvals[3] + lsetvals[0]
    if abs(c) < 1e-12:
        c = 0
    a = lsetvals[1] - lsetvals[0]
    b = lsetvals[3] - lsetvals[0]
    # print("a,b,c,d: ",a,b,c,d)
    xs = Symbol('xs')
    ys = Symbol('ys')
    levelset_py = d + a * xs + b * ys + c * xs * ys
    f_py = f(xs, ys)

    def levelset_pyl(x, y):
        return levelset_py.subs(ys, y).subs(xs, x)

    root1 = solve(levelset_py.subs(ys, 0), xs)
    root2 = solve(levelset_py.subs(ys, 1), xs)
    # print(root1, root2)
    part = [0, 1]
    for r in [root1[0], root2[0]]:
        if r > 0 and r < 1:
            part.append(r)
    part.sort()
    # print("Partitioning of [0,1]: ", part)

    referencevals = {}
    referencevals[POS] = 0
    referencevals[NEG] = 0
    if IF_INT:
        referencevals[IF] = 0

    def add_integration_on_interval(x0, x1):
        # print("Integrating on interval: ",x0,x1)
        y_ast = -(a * xs + d) / (b + c * xs)
        cut_left = levelset_pyl(x0, 0) * levelset_pyl(x0, 1) < 0
        cut_right = levelset_pyl(x1, 0) * levelset_pyl(x1, 1) < 0
        # print("Lset vals y = 0: ", levelset_pyl(x0,0), levelset_pyl(x1,0))
        # print("Lset vals y = 1: ", levelset_pyl(x0,1), levelset_pyl(x1,1))
        # print("Cuts: ", cut_left, cut_right)
        if (not cut_left) and (not cut_right):
            I0 = integrate(integrate(f_py, (ys, 0, 1)), (xs, x0, x1))
            if levelset_pyl(x0, 0) + levelset_pyl(x1, 0) > 0:
                print("Adding ", I, " to Part POS")
                referencevals[POS] += I0
            else:
                print("Adding ", I, " to Part NEG")
                referencevals[NEG] += I0
        else:
            I1 = integrate(integrate(f_py, (ys, 0, y_ast)), (xs, x0, x1))
            I2 = integrate(integrate(f_py, (ys, y_ast, 1)), (xs, x0, x1))
            Ia = integrate(f_py, (ys, 0, y_ast))
            # print("I1:", I1)
            # print("I2:", I2)

            # print("Interval length: ", x1-x0)
            # print("Sum of I1 and I2:", I1+I2)

            if(levelset_pyl(x0, 0) + levelset_pyl(x1, 0) > 0):
                referencevals[POS] += I1
                referencevals[NEG] += I2
            else:
                referencevals[POS] += I2
                referencevals[NEG] += I1

            if IF_INT:
                ts = Symbol('ts')
                C = Curve([ts, -(a * ts + d) / (b + c * ts)], (ts, x0, x1))
                print(C)
                referencevals[IF] += line_integrate(f_py, C, [xs, ys])

    for i in range(0, len(part) - 1):
        add_integration_on_interval(part[i], part[i + 1])
    return referencevals


if __name__ == "__main__":
    from netgen.geom2d import SplineGeometry
    square = SplineGeometry()
    square.AddRectangle([0, 0], [1, 1], bc=1)
    # mesh = Mesh (square.GenerateMesh(maxh=100, quad_dominated=False))
    mesh = Mesh(square.GenerateMesh(maxh=100, quad_dominated=True))

    lsetvals_list = [[-0.18687, 0.324987, 0.765764, 0.48983],
                     [0.765764, 0.324987, -0.18687, -0.48983],
                     [1, 2 / 3, -1, -2 / 3]]
    # lsetvals_list = [[1.,-1.,-3.,-1.]]
    # lsetvals_list = [[1,-1,-4,-2]]
    lsetvals_list.append([3, -1, 1, -1.023123])

    def f(x, y):
        return 1 + 0 * x + 0 * y

    f_ngs = f(x, y)
    V = H1(mesh, order=1)
    lset_approx = GridFunction(V)

    # domains = [NEG,POS, IF]
    domains = [NEG, POS]
    # domains = [IF]
    error_list = []

    max_order = 10
    # f1 = open("errors.dat","w")

    for lsetvals in lsetvals_list:
        print("Case lsetvals = ", lsetvals)
        referencevals = get_referencevals(lsetvals, f, IF in domains)
        print("referencevals: ", referencevals)
        levelset = get_levelset(lsetvals)

        InterpolateToP1(levelset, lset_approx)

        errors = dict()

        for key in domains:
            errors[key] = []
        inte = dict()

        for order in range(max_order + 1):
            errs = dict()
            for key in domains:
                lset_dom = {"levelset": lset_approx, "domain_type": key}
                integral = Integrate(levelset_domain=lset_dom,
                                     cf=f_ngs, mesh=mesh, order=order)
                inte[key] = integral
                print("PP: Integral on Domain ", key, " : ", integral)
                errors[key].append(abs(integral - referencevals[key]))
                errs[key] = abs(integral - referencevals[key])
            # f1.write(str(order) + "\t"
            #          + str(sqrt(pow(errs[POS],2.) + pow(errs[NEG],2.)))
            #          + "\n")
            # print("Sum of Part NEG, POS: ", inte[NEG] + inte[POS])
        print("L2 Errors:", errors)
        error_list.append(errors)
        # f1.write("\n\n")
        # Draw(levelset, mesh, "lset")
    print("All L2 Errors", error_list)



import pytest
from netgen.geom2d import SplineGeometry
from ngsolve import *
from xfem import *
from math import pi
from xfem.lset_spacetime import *

@pytest.mark.parametrize("imax", [5])
@pytest.mark.parametrize("order", [2,3,4])
def test_spacetime_lsetcurving_maxdist(imax, order):

    ngsglobals.msg_level = 1

    maxdist_at_reflevels = []
    # polynomial order in time
    k_t = order
    # polynomial order in space
    k_s = k_t

    for i in range(imax):
        square = SplineGeometry()
        square.AddRectangle([-0.6,-0.6],[0.6,0.6])
        ngmesh = square.GenerateMesh(maxh=0.5**(i+1), quad_dominated=False)
        mesh = Mesh (ngmesh)

        tstart = 0
        tend = 0.5
        delta_t = (tend - tstart)/(2**(i))

        told = Parameter(tstart)
        t = told + delta_t*tref

        lset_adap_st = LevelSetMeshAdaptation_Spacetime(mesh, order_space = k_s, order_time = k_t, threshold=0.5, discontinuous_qn=True)

        r0 = 0.5                                 

        rho =  CoefficientFunction((1/(pi))*sin(2*pi*t))
        r = sqrt(x**2+(y-rho)**2)
        levelset= CoefficientFunction(r -r0)

        maxdists = []
        while tend - told.Get() > delta_t/2:
            dfm = lset_adap_st.CalcDeformation(levelset)
            mesh.deformation = dfm
            maxdist = lset_adap_st.CalcMaxDistance(levelset)    
            mesh.deformation = None
            maxdists.append(maxdist)
            told.Set(told.Get() + delta_t)

        print("i: ", i, "\tdist max: ", max(maxdists))
        maxdist_at_reflevels.append(max(maxdists))

    eocs = [log(maxdist_at_reflevels[i-1]/maxdist_at_reflevels[i])/log(2) for i in range(1,len(maxdist_at_reflevels))]
    print("eocs: ", eocs)
    avg_eoc = sum(eocs[-2:])/len(eocs[-2:])
    print("avg: ", avg_eoc)

    assert avg_eoc > order + 0.75

if __name__ == "__main__":
    test_spacetime_lsetcurving_maxdist(5, 2)


"""
This example (visually) illustrates the usage of the usage and
functionality of xfem.mlset to generate domain descriptions based on
multiple, simpler (shorter) multi-level set domain descriptions.
"""

# ------------------------------ LOAD LIBRARIES -------------------------------
from netgen.geom2d import SplineGeometry
from ngsolve import *
from xfem import *
from xfem.mlset import *


# ----------------------------------- MAIN ------------------------------------
square = SplineGeometry()
square.AddRectangle((-2.5, -1), (2.5, 2.5), bc=1)
mesh = Mesh(square.GenerateMesh(maxh=1, quad_dominated=False))


# Separate regions
level_sets1 = [-2*x + y - 1, 2*x + y - 1, y - 2]
dta1 = DomainTypeArray((POS, POS, NEG))
Draw(dta1.Indicator(level_sets1), mesh, "dta1", sd=4)

level_sets2 = [-2*x + y - 4, 2*x + y + 2, y]
dta2 = DomainTypeArray((NEG, NEG, POS))
Draw(dta2.Indicator(level_sets2), mesh, "dta2", sd=4)

level_sets3 = [x - 2*y - 1, x + 2*y - 1, x - 2]
dta3 = DomainTypeArray((POS, POS, NEG))
Draw(dta3.Indicator(level_sets3), mesh, "dta3", sd=4)


# Test TensorUnion
dta_union = TensorUnion(dta1, dta2, dta3)
Draw(dta_union.Indicator(level_sets1 + level_sets2 + level_sets3),
     mesh, "dta_union", sd=5)

dta4 = DomainTypeArray([(POS, POS, NEG, ANY, ANY, ANY),
                        (ANY, ANY, ANY, NEG, NEG, POS)])
dta_union2 = TensorUnion(dta4, dta3)
Draw(dta_union.Indicator(level_sets1 + level_sets2 + level_sets3),
     mesh, "dta_union2", sd=5)


# Test TensorIntersection
dta5 = ~ dta1
dta6 = ~ dta2
dta7 = ~ dta3

dta_intersection = TensorIntersection(dta5, dta6, dta7)
Draw(dta_intersection.Indicator(level_sets1 + level_sets2 + level_sets3),
     mesh, "dta_intersection", sd=5)

bnd1 = dta1.Boundary()
dta_intersection2 = TensorIntersection(bnd1, dta6)
Draw(dta_intersection2.IndicatorSmoothed(level_sets1 + level_sets2),
     mesh, "dta_intersection2", sd=6)



# solve the Poisson equation -Delta u = f
# with Dirichlet boundary condition u = 0

from ngsolve import *
from ngsolve.krylovspace import CG
from netgen.geom2d import unit_square
from netgen.geom2d import SplineGeometry
from xfem import *

ngsglobals.msg_level = 1

# generate a triangular mesh of mesh-size 0.2
square = SplineGeometry()
square.AddRectangle([-1.5,-1.5],[1.5,1.5],bc=1)
mesh = Mesh (square.GenerateMesh(maxh=0.2, quad_dominated=False))

mu = [1e-0, 1e-7]
order = 1


rhsscal = -4*mu[0]*mu[1] 
coef_f = [ rhsscal, rhsscal ]
distsq = x*x + y*y - 1
solution = [ mu[1]* distsq, mu[0] * distsq ]

levelset = ( sqrt(x*x + y*y) - 1 )
subdivlvl = 0

def main():
    print('hallo')
    gamma_stab = 10
    nref = 2

    params =    {  
                    "order"     : order,
                    "mu"        : mu,
                    "gamma_stab": gamma_stab,
                    "kappa"     : "harmonic" # hansbo, highorder
                }


    # levelset fe space
    VhL = H1(mesh, order = order, dirichlet=[], dgjumps = True)
    lsetp1 = GridFunction(VhL)
    InterpolateToP1(levelset,lsetp1)
    ci = CutInfo(mesh, lsetp1)   
    
    # cut fe space
    Vh = H1(mesh, order=order, dirichlet=[1,2,3,4])
    VhNeg = Compress( Vh, active_dofs = GetActiveDof(mesh, HASNEG) )
    VhPos = Compress( Vh, active_dofs = GetActiveDof(mesh, HASPOS) )

    CutVh = FESpace( [VhNeg, VhPos], flags={"dgjumps": True} )

    # construct prolongation for neg/pos space
    prolongNeg = P1Prolongation(mesh)
    prolongPos = P1Prolongation(mesh)
    # create cut prolongation
    CutProl = CompoundProlongation( CutVh )
    CutProl.AddProlongation(prolongNeg)
    CutProl.AddProlongation(prolongPos)
    prolongNeg.Update(VhNeg)
    prolongPos.Update(VhPos)
    

    #create coarse level bilinear/linear form
    a,f = AssembleCutPoisson( CutVh=CutVh, ci=ci, lsetp1=lsetp1, params=params )

    
    #list of coarse grid matrices
    mats = {}
    mats[0] = a.mat
    fvecs = {}
    fvecs[0] = f.vec
    #list of smoothers
    smoothers = {}
    smoothers[0] = CutFemSmoother(a=a, CutVh=CutVh, ci=ci) 
    #coarse grid solver
    inv = a.mat.Inverse(CutVh.FreeDofs(), inverse="sparsecholesky")
    

    # create multilevel hierarchy
    for i in range(nref):
        mesh.Refine()
        #levelset update
        VhL.Update()
        lsetp1.Update()
        InterpolateToP1(levelset,lsetp1)
        ci = CutInfo(mesh,lsetp1)
        #cut space update
        Vh.Update()
        CutVh.components[0].SetActiveDofs( GetActiveDof(mesh,HASNEG) )
        CutVh.components[1].SetActiveDofs( GetActiveDof(mesh,HASPOS) )
        CutVh.Update()
        CutProl.Update(CutVh)
        #reassemble
        a,f = AssembleCutPoisson( CutVh=CutVh, ci=ci, lsetp1=lsetp1, params=params )        
        mats[i+1] = a.mat
        fvecs[i+1] = f.vec

        smoothflag = (i == nref-1) #or (i==nref-2)
        smoothers[i+1] = CutFemSmoother(a=a, CutVh=CutVh, ci=ci, finest=smoothflag) 

    

    # create multigrid solver
    MGpre = MultiGridCL ( level=nref, prol=CutProl, 
                          matrices=mats, smoothers=smoothers,
                          coarsegridsolver=inv, f=f, tau=2 
                        )

    # start (richardson) iteration process with multigrid preconditioner
    usol = GridFunction(CutVh) 
    usol.vec[:] = 0.

    usol.vec.Data = NestedIteration( MGpre, fvecs, usol.vec, nref )

    res = usol.vec.CreateVector()
    projres = usol.vec.CreateVector()
    tol = 1e-6
    proj = Projector(mask=CutVh.FreeDofs(),range=True) 
    fnew = f.vec.data     

    res.data = fnew - a.mat*usol.vec
    projres.data = proj*res
    #projres.data = proj*fnew #f.vec
    normf = Norm(projres)

    oldres = normf

    print( "norm rhs: {0:.2E}".format(normf) )

    for it in range(1,50):
        usol.vec.data = MGpre*fnew
        res.data = fnew - a.mat*usol.vec
        projres.data = proj*res
        res_norm = Norm(projres) #/ normf
        #print("it =", it, " ||res||_2 =", res_norm)
        print("mg-it =", it, "\t ||res||_2 = {0:.2E}".format(res_norm), "\t reduction: {0:.2f}".format(res_norm/oldres))
        oldres = res_norm
        if res_norm < tol*normf:
            break

    # input('w')

    udraw = GridFunction( CutVh )
    udraw.components[1].Set( solution[1], BND )
    udraw.vec.data += usol.vec

    # input('w1')

    # Computation of L2 error:
    err_sqr_coefs = [(udraw.components[i]-solution[i])*(udraw.components[i]-solution[i]) for i in [0,1] ]
    lset_doms = LsetDoms( levelset, lsetp1, subdivlvl )
    # input('w2')
    #l2error = sqrt( sum( [Integrate(lset_doms[i], cf=err_sqr_coefs[i], mesh=mesh, order=2*order, heapsize=1000000) for i in [0,1] ]))   
    l2error = sqrt( sum( [Integrate(lset_doms[i], cf=err_sqr_coefs[i], mesh=mesh) for i in [0,1] ]))   
    print("L2 error : ",l2error)

    input('draw?\n')

    # if( nref > 2):
    #     print('nref too large')
    #     return
    #     print('test')

    print('drawing')
    # pretty printing    
    ucoef = IfPos( lsetp1, udraw.components[1], udraw.components[0] )
    Draw (ucoef,mesh,'u')

    input('finish? - seg fault\n')
    del MGpre
    

def GetActiveDof(mesh,HASPOSNEG):
        Vh = H1(mesh, order = order, dirichlet=[], dgjumps = True)
        lsetp1 = GridFunction(Vh)
        InterpolateToP1(levelset,lsetp1)
        ci = CutInfo(mesh,lsetp1)
        return GetDofsOfElements(Vh,ci.GetElementsOfType(HASPOSNEG))


def LsetDoms( levelset, lsetp1, subdivlvl ):
    if (subdivlvl == 0):
        lsetint = lsetp1
    else:
        lsetint = levelset

    lset_neg = { "levelset" : lsetint, "domain_type" : NEG, "subdivlvl" : subdivlvl}
    lset_pos = { "levelset" : lsetint, "domain_type" : POS, "subdivlvl" : subdivlvl}
    lset_if  = { "levelset" : lsetint, "domain_type" : IF , "subdivlvl" : subdivlvl}

    lset_dom = [ lset_neg, lset_pos,  lset_if ]

    return lset_dom

def KappaLambda(params,ci):
    kappac  = params["kappa"]
    mu      = params["mu"]
    order   = params["order"]

    if kappac == "harmonic":
        kappa = [ mu[1] / ( mu[0] + mu[1] ) , mu[0] /  ( mu[0] + mu[1] ) ]
        # stabilization parameter for Nitsche
        lambda_nitsche  = 2*mu[0]*mu[1]/(mu[0] + mu[1]) * 10 * order * order
    elif kappac == "hansbo":
        kappa = [CutRatioGF(ci),1.0-CutRatioGF(ci)]
        # stabilization parameter for Nitsche
        lambda_nitsche  = 0.5 * (mu[0]+mu[1]) * 20 * (order+1) * order
    elif kappac == "highorder":
        kappaHO = IfPos( CutRatioGF(ci) -0.5, 1, 0 )
        kappa = [kappaHO,1.0-kappaHO]
        # stabilization parameter for Nitsche
        lambda_nitsche  = 0.5 * (mu[0]+mu[1]) * 20 * order * order

    return [ kappa, lambda_nitsche ]

def AssembleCutPoisson(**kwargs):
    ci = kwargs['ci']
    hasneg      = ci.GetElementsOfType(HASNEG)  # <- "hasneg": has (also) negative level set values
    haspos      = ci.GetElementsOfType(HASPOS)  # <- "haspos": has (also) positive level set values    
    hasif       =  BitArray(ci.GetElementsOfType(IF))
    ba_facets   = [GetFacetsWithNeighborTypes(mesh,a=hasneg,b=hasif),GetFacetsWithNeighborTypes(mesh,a=haspos,b=hasif)]


    lsetp1 = kwargs['lsetp1']
    n_lset  = 1.0/grad(lsetp1).Norm() * grad(lsetp1)
    h       = specialcf.mesh_size
    params = kwargs['params']
    [kappa, lambda_nitsche ] = KappaLambda(params,ci)    

    # expressions of test and trial functions (u and v are tuples):
    CutVh = kwargs['CutVh']
    (u,v) = CutVh.TnT()

    gradu = [grad(ui) for ui in u]
    gradv = [grad(vi) for vi in v]

    mu = params["mu"]
    average_flux_u = sum([- kappa[i] * mu[i] * gradu[i] * n_lset for i in [0,1]])
    average_flux_v = sum([- kappa[i] * mu[i] * gradv[i] * n_lset for i in [0,1]])

    # integration domains
    lset_doms = LsetDoms( levelset, lsetp1, subdivlvl )
    
    # bilinear forms:
    a = BilinearForm(CutVh, symmetric = False)
    f = LinearForm(CutVh)

    #domain integrals:
    for i in [0,1] :
        a += SymbolicBFI( lset_doms[i], form = mu[i] * gradu[i] * gradv[i] )        

        #rhs
        f += SymbolicLFI( lset_doms[i], form = coef_f[i] * v[i] )

    # Nitsche integrals:
    a += SymbolicBFI(lset_doms[2], form = average_flux_u * (v[0]-v[1])
                                        + average_flux_v * (u[0]-u[1])
                                        + lambda_nitsche / h * (u[0]-u[1]) * (v[0]-v[1]) ) 

    for i in [0,1]:
        # Ghost penalty
        a += SymbolicFacetPatchBFI( form =  params["gamma_stab"]  * 
                                    mu[i] / h / h * 
                                    (u[i] - u[i].Other()) * (v[i] - v[i].Other()), 
                                    skeleton=False, definedonelements=ba_facets[i] )

    # # apply mesh adaptation 
    # if( doDeform ):   
    #     mesh.SetDeformation(deformation)

    # setting up matrix and vector
    a.Assemble()
    f.Assemble()

    #updating with non-homogeneous boundary conditions
    gfu = GridFunction( CutVh )
    #boundary lies in positive part
    gfu.components[1].Set( solution[1], BND )
    f.vec.data -= a.mat * gfu.vec

    return (a,f)

def IfaceUnks(CutVh, ci):
    ifdofs = BitArray(CutVh.ndof)
    ifdofs[:] = False
    hasif = BitArray( ci.GetElementsOfType(IF) )

    for el in CutVh.Elements():
        if( hasif[el.nr] ):
            verts = el.vertices
            for v in verts:
                # print( CutVh.GetDofNrs(v) )
                # print( CutVh.components[0].GetDofNrs(v) )
                # print( CutVh.components[1].GetDofNrs(v) )
                # print( CutVh.components[0].ndof )
                # print('-------------------------------')
                for unk in CutVh.GetDofNrs(v):                    
                    ifdofs[unk] = True

    ifdofs &= CutVh.FreeDofs()
    #print(ifdofs)
    #input('weiter')
    return ifdofs

class CutFemSmoother:     
    def __init__(self,**kwargs):
        self.a = kwargs["a"]
        ifdofs = IfaceUnks(kwargs["CutVh"],kwargs["ci"])
        #self.proj = Projector(mask=self.ifdofs,range=True)
        ifdofslst = [ [i] for i in range(len(ifdofs)) if ifdofs[i]==True ]
        self.proj = self.a.mat.CreateBlockSmoother(ifdofslst)
        #self.inv =  self.a.mat.Inverse(ifdofs, inverse="sparsecholesky")
        self.preJ = self.a.mat.CreateSmoother( kwargs["CutVh"].FreeDofs() )
        self.finest = kwargs.get('finest',False)
    def __del__(self):
            print('smoother dying')
            input('destruktor')
    def IfCorr( self, u, rhs):
        #if not self.finest:
        #    return
        update = u.CreateVector()
        # this needs to be more efficient 
        # with help of ifdofs ...
        update.data = self.a.mat * u - rhs
        cgoutput = u.CreateVector()
        cgoutput.data = CG(self.a.mat, update, pre=self.proj, tol=1e-4,printrates=False)
        #u.data -= self.inv * update
        u.data -= cgoutput

    def Smooth( self, u,rhs, nu ):
        for k in range(nu):
            self.preJ.Smooth(u, rhs)
        self.IfCorr(u,rhs)
        
        

class MultiGridCL(BaseMatrix):
        def __init__(self,**kwargs):
            super(MultiGridCL,self).__init__()
            self.nu     = kwargs.get('nu',2)            
            self.l      = kwargs['level']
            self.prol   = kwargs['prol']
            self.mats   = kwargs['matrices']
            self.smoothers  = kwargs['smoothers']
            self.inv    = kwargs["coarsegridsolver"]
            f = kwargs["f"]
            nref = self.l
            self.errors     = [f.vec.CreateVector() for i in range(nref+1)]
            self.defects    = [f.vec.CreateVector() for i in range(nref+1)]

            self.cycles = kwargs.get('tau',1) #v-cycle per default
        
        def __del__(self):
            print('mg dying')
            input('destruktor')
            del self.mats
            print('somehow this is crashing ...')
            del self.prol
            del self.smoothers
            


        def MultiGridIter(self,rhs,u,level):
            error   = self.errors[level]
            defect  = self.defects[level]

            if level == 0 :
                u.data = self.inv*rhs
                return
        #smoothing
            #for i in range(self.nu):
            self.smoothers[level].Smooth(u,rhs,self.nu)
        #compute defect
            defect.data = rhs - self.mats[level]*u
        #restrict defect
            self.prol.Restrict(level,defect)
            error[:] = 0
            #v-cycle
            for k in range( self.cycles ):
                self.MultiGridIter(defect,error,level-1)
            self.prol.Prolongate(level,error)
            u.data += error
        #smoothing
            #for i in range(self.nu):
            self.smoothers[level].Smooth(u,rhs,self.nu)
            return u

        def Mult(self,w,z):
            #z[:] = 0.0
            #z.data = w
            #z.data = 
            self.MultiGridIter(w,z,self.l)
        def Height(self):
            return a.mat.height
        def Width(self):
            return a.mat.height

def NestedIteration( mgcl, fvecs, u0, lastlevel):
    u0.Data = mgcl.inv * fvecs[0]

    for i in range(1,lastlevel+1):
        mgcl.prol.Prolongate(i,u0)
        ftmp = u0.CreateVector()
        ftmp[:] = 0.
        ftmp.Range( 0, len(fvecs[i]) ).data = fvecs[i]        
        u0.Data = mgcl.MultiGridIter(ftmp,u0,i)

    return u0

if __name__ == '__main__':
    main()   



"""
In this example we solve a scalar *unfitted* PDE problem on a moving
domain. A discontinuous-in-time space-time formulation is applied.
Natural boundary conditions are applied which simplifies the variational
formulation. To stabilize arbitrary cut configurations, we use a
space-time version of the ghost penalty method.

Domain:
-------
The background domain is [-0.6,0.6]x[-1,1]x[0,0.5] (2D+time interval)
while the physical domain is a circle that is traveling up and down
over time.

PDE problem:
------------
  u_t + wx·u_x + wy·u_y - (u_xx + u_yy) = f in  Omega(t) (where lset is neg.)
        (-u_x+wx u)·nx+(-u_y+wy u)·ny u = 0 on dOmega(t) (where lset is zero.)
where w = (wx,wy) is a divergence-free vector field.
The r.h.s. term f is chosen according to a manufactured solution.

Discretisation:
---------------
* Background space-time finite element space restricted to active domain

* Ghost penalty stabilization to deal with bad cuts (version as in [1])

Implementational aspects (cf. [1] and [2] for details):
-------------------------------------------------------
* Geometry approximation in space-time using isoparametric unfitted FEM

* Projection operator that maps solutions from one deformed mesh to another

* A (sparse) direct solver is applied to solve the arising linear systems.

References:
-----------
All concepts that are used here are explained in the jupyter-tuorials
`spacetime.ipynb`. As a simplified setting without cut configurations,
we also refer to the `spacetimeDG_fitted.py` demo.

Literature:
-----------
[1] J. Preuß, Higher order unfitted isoparametric space-time FEM on moving
    domains. Master's thesis, NAM, University of Göttingen, 2018.
[2] F. Heimann. On Discontinuous- and Continuous-In-Time Unfitted Space-Time
    Methods for PDEs on Moving Domains. Master's thesis, NAM, University of
    Göttingen, 2020.
"""

# ------------------------------ LOAD LIBRARIES -------------------------------
from ngsolve import *
from ngsolve.meshes import OrthoBrick, Pnt, CSGeometry
from xfem import *
from math import pi
from xfem.lset_spacetime import *
ngsglobals.msg_level = 1

# -------------------------------- PARAMETERS ---------------------------------
# DISCRETIZATION PARAMETERS:

# Parameter for refinement study:
i = 2
n_steps = 2**i
space_refs = i

# Polynomial order in time
k_t = 2
# Polynomial order in space
k_s = k_t
# Polynomial order in time for level set approximation
lset_order_time = k_t
# Integration order in time
time_order = 2 * k_t
# Time stepping parameters
tstart = 0
tend = 0.5
delta_t = (tend - tstart) / n_steps
maxh = 0.5
# Ghost-penalty parameter
gamma = 0.05
# Map from reference time to physical time
told = Parameter(tstart)
t = told + delta_t * tref

# PROBLEM SETUP:
geometry = CSGeometry()
geometry.Add(OrthoBrick(Pnt(-0.6, -1, -0.6), Pnt(0.6, 1, 0.6)))

D = 3

(xmin, xmax) = (-0.6, 0.6)
(ymin, ymax) = (-0.85, 0.85)
(zmin, zmax) = (-0.6, 0.6)

# Level set geometry
# Radius of sphere (the geometry)
R = 0.5
# Position shift of the geometry in time
rho = (1 / (pi)) * sin(2 * pi * t)
# Convection velocity:
w = CoefficientFunction((0, rho.Diff(t), 0))
max_velocity = 2

# Level set
r = sqrt(x**2 + (y - rho)**2 + z**2)
levelset = r - R

# Diffusion coefficient
alpha = 1
# Solution
u_exact = cos(pi * r / R) * sin(pi * t)
# R.h.s.
coeff_f = (u_exact.Diff(t)
           - alpha * (u_exact.Diff(x).Diff(x) + u_exact.Diff(y).Diff(y)
                      + u_exact.Diff(z).Diff(z))
           + w[0] * u_exact.Diff(x) + w[1] * u_exact.Diff(y)
           + w[2] * u_exact.Diff(z)).Compile()

# ----------------------------------- MAIN ------------------------------------
ngmesh = geometry.GenerateMesh(maxh=maxh, quad_dominated=False)
for j in range(space_refs):
    ngmesh.Refine()
mesh = Mesh(ngmesh)

# Spatial FESpace for solution
fes1 = L2(mesh, order=k_s, dgjumps=True)
# Time finite element (nodal!)
tfe = ScalarTimeFE(k_t)
# (Tensor product) space-time finite element space
st_fes = tfe * fes1

# Space time version of Levelset Mesh Adapation object. Also offers integrator
# helper functions that involve the correct mesh deformation
lsetadap = LevelSetMeshAdaptation_Spacetime(mesh, order_space=k_s,
                                            order_time=lset_order_time,
                                            threshold=0.5,
                                            discontinuous_qn=True)

gfu = GridFunction(st_fes)
u_last = CreateTimeRestrictedGF(gfu, 1)

scene = DrawDC(lsetadap.levelsetp1[TOP], u_last, 0, mesh, "u_last",
               deformation=lsetadap.deformation[TOP])

u, v = st_fes.TnT()
h = specialcf.mesh_size

ba_facets = BitArray(mesh.nfacet)
ba_facets_inner = BitArray(mesh.nfacet)
ci = CutInfo(mesh, time_order=0)

dQ = delta_t * dCut(lsetadap.levelsetp1[INTERVAL], NEG, time_order=time_order,
                    deformation=lsetadap.deformation[INTERVAL],
                    definedonelements=ci.GetElementsOfType(HASNEG))
dQ_f = delta_t * dCut(lsetadap.levelsetp1[INTERVAL], NEG,
                      time_order=time_order,
                      deformation=lsetadap.deformation[INTERVAL],
                      definedonelements=ba_facets_inner, skeleton=True)
dOmold = dCut(lsetadap.levelsetp1[BOTTOM], NEG,
              deformation=lsetadap.deformation[BOTTOM],
              definedonelements=ci.GetElementsOfType(HASNEG), tref=0)
dOmnew = dCut(lsetadap.levelsetp1[TOP], NEG,
              deformation=lsetadap.deformation[TOP],
              definedonelements=ci.GetElementsOfType(HASNEG), tref=1)
dw = delta_t * dFacetPatch(definedonelements=ba_facets, time_order=time_order,
                           deformation=lsetadap.deformation[INTERVAL])


def dt(u):
    return 1.0 / delta_t * dtref(u)


jump_u = u-u.Other()
jump_v = v-v.Other()
n = specialcf.normal(D)
mean_dudn = 0.5*n * (grad(u)+grad(u).Other())
mean_dvdn = 0.5*n * (grad(v)+grad(v).Other())

a = RestrictedBilinearForm(st_fes, "a", check_unused=False,
                           element_restriction=ci.GetElementsOfType(HASNEG),
                           facet_restriction=ba_facets_inner)
a += v * (dt(u) - dt(lsetadap.deform) * grad(u)) * dQ
a += (alpha * InnerProduct(grad(u), grad(v))) * dQ
a += (v * InnerProduct(w, grad(u))) * dQ
a += u * v * dOmold

a += (-mean_dvdn*jump_u - mean_dudn*jump_v + 50*k_s**2/h*jump_u*jump_v) * dQ_f

a += h**(-2) * (1 + delta_t / h) * gamma * \
    (u - u.Other()) * (v - v.Other()) * dw

f = LinearForm(st_fes)
f += coeff_f * v * dQ
f += u_last * v * dOmold

# Set initial values
u_last.Set(fix_tref(u_exact, 0))
# Project u_last at the beginning of each time step
lsetadap.ProjectOnUpdate(u_last)

while tend - told.Get() > delta_t / 2:
    lsetadap.CalcDeformation(levelset)

    # Update markers in (space-time) mesh
    ci.Update(lsetadap.levelsetp1[INTERVAL], time_order=0)

    # re-compute the facets for stabilization:
    ba_facets[:] = GetFacetsWithNeighborTypes(mesh,
                                              a=ci.GetElementsOfType(HASNEG),
                                              b=ci.GetElementsOfType(IF))
    ba_facets_inner[:] = GetFacetsWithNeighborTypes(
                                        mesh,
                                        a=ci.GetElementsOfType(HASNEG),
                                        b=ci.GetElementsOfType(HASNEG))
    active_dofs = GetDofsOfElements(st_fes, ci.GetElementsOfType(HASNEG))

    a.Assemble(reallocate=True)
    f.Assemble()

    # Solve linear system
    inv = a.mat.Inverse(active_dofs, inverse="")
    gfu.vec.data = inv * f.vec.data

    # Evaluate upper trace of solution for
    #  * for error evaluation
    #  * upwind-coupling to next time slab
    RestrictGFInTime(spacetime_gf=gfu, reference_time=1.0, space_gf=u_last)

    # Compute error at final time
    l2error = sqrt(Integrate((u_exact - u_last)**2 * dOmnew, mesh))

    # Update time variable (ParameterCL)
    told.Set(told.Get() + delta_t)
    print("\rt = {0:12.9f}, L2 error = {1:12.9e}".format(told.Get(), l2error))

    try:
        __builtin__
        __IPYTHON__
        scene.Redraw()
    except NameError:
        scene.Redraw(blocking=True)



"""
	Description:
	------------
	Sharp cube defined by six level sets, one for each face.
"""

# Import libraries
from netgen.csg import *
from ngsolve import *
from xfem import *
from xfem.mlset import *

# Mesh
geo = CSGeometry()
geo.Add(OrthoBrick(Pnt(-0.8,-0.8,-0.8), Pnt(0.8,0.8,0.8)))
mesh = Mesh(geo.GenerateMesh(maxh=1))

# -----------------------------------------------------------------------------
# First Test

# Level sets
level_sets = (x - 0.5, x + 0.5, x - y, z - 0)
nr_ls = len(level_sets)
level_sets_p1 = tuple(GridFunction(H1(mesh,order=1)) for i in range(nr_ls))

for i, lset_p1 in enumerate(level_sets_p1):
    InterpolateToP1(level_sets[i], lset_p1)

# Define Domain
line = DomainTypeArray(dtlist=[(NEG, POS, IF, IF)])
Draw(line.IndicatorSmoothed(level_sets_p1), mesh, "line", sd=6)

# Compute length
length = Integrate(levelset_domain={"levelset": level_sets_p1,
                                    "domain_type": line},
                   mesh=mesh, cf=1, order=0)

print("length = {:10.8f}".format(length))
print("length error = {:4.3e}".format(abs(length - sqrt(2))))
assert abs(length - sqrt(2)) < 1e-12

del level_sets, level_sets_p1, nr_ls, line

# -----------------------------------------------------------------------------
# Second Test: cross

# Level sets
level_sets = (x - 0.5, - x - 0.5, z - y, x - z, x + z)
nr_ls = len(level_sets)
level_sets_p1 = tuple(GridFunction(H1(mesh,order=1)) for i in range(nr_ls))

for i, lset_p1 in enumerate(level_sets_p1):
    InterpolateToP1(level_sets[i], lset_p1)

# Define Domain
cross = DomainTypeArray([(NEG, NEG, IF, IF, ANY), (NEG, NEG, IF, ANY, IF)])
Draw(cross.IndicatorSmoothed(level_sets_p1), mesh, "cross", sd=6)

# Compute length
length = Integrate(levelset_domain={"levelset": level_sets_p1,
                                    "domain_type": cross},
                   mesh=mesh, cf=1, order=0)

print("length = {:10.8f}".format(length))
print("length error = {:4.3e}".format(abs(length - 2*sqrt(3))))
assert abs(length - 2*sqrt(3)) < 1e-12

del level_sets, level_sets_p1, nr_ls, cross

# -----------------------------------------------------------------------------
# Third test: Parallel lines close to each other

level_sets = (x - 0.5, - x - 0.5, - z - 0.5, z + y, x + y, z + y + 0.1)
nr_ls = len(level_sets)
level_sets_p1 = tuple(GridFunction(H1(mesh,order=1)) for i in range(nr_ls))

for i, lset_p1 in enumerate(level_sets_p1):
    InterpolateToP1(level_sets[i], lset_p1)

lines = DomainTypeArray([(NEG, NEG, NEG, IF, IF, ANY), 
                         (NEG, NEG, NEG, ANY, IF, IF)])

length = Integrate(levelset_domain={"levelset": level_sets_p1,
                                    "domain_type": lines},
                   mesh=mesh, cf=1, order=0)

print("length = {:10.8f}".format(length))
print("length error = {:4.3e}".format(abs(length - 1.9*sqrt(3))))
assert abs(length - 1.9*sqrt(3)) < 1e-12


"""
example of a triangle described by three level set functions
"""

# the constant pi
from math import pi
# ngsolve stuff
from ngsolve import *
# basic xfem functionality
from xfem import *
# basic geometry features (for the background mesh)
from netgen.geom2d import SplineGeometry

# We generate the background mesh of the domain and use a simplicial triangulation
# To obtain a mesh with quadrilaterals use 'quad_dominated=True'

square = SplineGeometry()
square.AddRectangle([-1,-0.5],[1,1.5],bc=1)
mesh = Mesh (square.GenerateMesh(maxh=0.2, quad_dominated=False))

lsetp1_a = GridFunction(H1(mesh,order=1))
lsetp1_b = GridFunction(H1(mesh,order=1))
lsetp1_c = GridFunction(H1(mesh,order=1))

InterpolateToP1(     y-1,lsetp1_a)
InterpolateToP1( 2*x-y  ,lsetp1_b)
InterpolateToP1(-2*x-y  ,lsetp1_c)

Draw (lsetp1_a, mesh, "lset_a")
Draw (lsetp1_b, mesh, "lset_b")
Draw (lsetp1_c, mesh, "lset_c")

Draw (lsetp1_a * lsetp1_b * lsetp1_c, mesh, "lset_mult")

inner = { "levelset" : (lsetp1_a,lsetp1_b,lsetp1_c),
          "domain_type" : (NEG,NEG,NEG)}

boundary = { "levelset" : (lsetp1_a,lsetp1_b,lsetp1_c),
             "domain_type" : [(IF,NEG,NEG), (NEG,IF,NEG), (NEG,NEG,IF)]}

# outer = { "levelsets" : (lsetp1_a,lsetp1_b,lsetp1_c),
#           "domain_type" : ~(NEG,NEG,NEG)}

triangle_area = Integrate( levelset_domain=inner, cf=1, mesh=mesh, order=2)
triangle_boundary_length = Integrate( levelset_domain=boundary, cf=1, mesh=mesh, order=2)
# outer_area = Integrate( levelset_domain=outer, cf=1, mesh=mesh, order=2)



"""
example of a triangle described by three level set functions
"""

from netgen.geom2d import SplineGeometry
from ngsolve import *
from xfem import *
from xfem.mlset import *


# Background mesh
square = SplineGeometry()
square.AddRectangle([-1, -0.5], [1, 1.5], bc=1)
mesh = Mesh(square.GenerateMesh(maxh=0.2, quad_dominated=False))

# Level sets
lsetp1_a = GridFunction(H1(mesh, order=1))
lsetp1_b = GridFunction(H1(mesh, order=1))
lsetp1_c = GridFunction(H1(mesh, order=1))

InterpolateToP1(y - 1, lsetp1_a)
InterpolateToP1(2 * x - y, lsetp1_b)
InterpolateToP1(-2 * x - y, lsetp1_c)

Draw(lsetp1_a, mesh, "lset_a")
Draw(lsetp1_b, mesh, "lset_b")
Draw(lsetp1_c, mesh, "lset_c")

# Shape of interest
triangle = DomainTypeArray([(NEG, NEG, NEG)])
Draw(triangle.Indicator((lsetp1_a, lsetp1_b, lsetp1_c)), mesh, "triangle_dom")
Draw(triangle.Boundary().IndicatorSomoothed((lsetp1_a, lsetp1_b, lsetp1_c)), 
                                            mesh, "triangle_bnd")

# Show MultiLevelsetCutInfo functionality
mlci = MultiLevelsetCutInfo(mesh, (lsetp1_a, lsetp1_b, lsetp1_c))

els_neg = mlci.GetElementsOfType(triangle)
els_neg2 = mlci.GetElementsOfType(triangle.as_list)
els_not_neg = mlci.GetElementsOfType(~triangle)
els_if = mlci.GetElementsWithContribution(triangle.Boundary())
els_hasneg = mlci.GetElementsWithContribution(triangle)
els_hasneg2 = mlci.GetElementsWithContribution(triangle.as_list)

els_if2 = BitArray(mesh.ne)
els_if2[:] = False
els_if2 |= els_hasneg & ~els_neg

# Draw BitArrays
Draw(BitArrayCF(els_neg), mesh, "els_neg")
Draw(BitArrayCF(els_neg2), mesh, "els_neg2")
Draw(BitArrayCF(els_hasneg), mesh, "els_hasneg")
Draw(BitArrayCF(els_hasneg2), mesh, "els_hasneg2")
Draw(BitArrayCF(els_if), mesh, "els_if")
Draw(BitArrayCF(els_if2), mesh, "els_if2")



"""
In this example we integrate over a quarter circle on a series of
quadrilateral meshes.
"""


# ------------------------------ LOAD LIBRARIES -------------------------------
from netgen.geom2d import SplineGeometry
from ngsolve import *
from xfem import *
from xfem.lsetcurv import *
from math import pi


# -------------------------------- PARAMETERS ---------------------------------
# Radius of circle
r = 0.6
# Number of mesh refinements
n_ref = 8
# Integration order
order = 2


# ----------------------------------- MAIN ------------------------------------

square = SplineGeometry()
square.AddRectangle([0, 0], [1, 1], bc=1)
# mesh = Mesh (square.GenerateMesh(maxh=100, quad_dominated=False))
mesh = Mesh(square.GenerateMesh(maxh=100, quad_dominated=True))

# domains = [NEG,POS,IF]
domains = [NEG, POS, IF]

levelset = sqrt(x * x + y * y) - r
referencevals = {POS: 1 - pi * r * r / 4, NEG: pi * r * r / 4, IF: r * pi / 2}


errors = dict()
eoc = dict()

for key in domains:
    errors[key] = []
    eoc[key] = []

for i in range(n_ref):
    lsetmeshadap = LevelSetMeshAdaptation(mesh, order=order, threshold=0.2,
                                          discontinuous_qn=True)
    V = H1(mesh, order=1)
    lset_approx = GridFunction(V)
    InterpolateToP1(levelset, lset_approx)
    # Draw(lset_approx,mesh,"lset_p1")

    f = CoefficientFunction(1)

    deformation = lsetmeshadap.CalcDeformation(levelset)
    # mesh.SetDeformation(deformation)
    # Draw(deformation,mesh,"deformation")

    for key in domains:
        lset_dom = {"levelset": lset_approx, "domain_type": key}
        integral = Integrate(levelset_domain=lset_dom,
                             cf=f, mesh=mesh, order=order)
        print("Result of Integration Reflevel ",
              i, ", Key ", key, " : ", integral)
        errors[key].append(abs(integral - referencevals[key]))

    mesh.deformation = None

    if i < n_ref - 1:
        # RefineAtLevelSet(gf=lset_approx)
        mesh.Refine()

for key in domains:
    eoc[key] = [log(errors[key][i + 1] / errors[key][i]) / log(0.5)
                for i in range(n_ref - 1)]

print("L2-errors:", errors)
# l2_eoc = [log(errors[i+1]/errors[i])/log(0.5) for i in range(n_ref-1)]
print("experimental order of convergence (L2):", eoc)

for key in domains:
    for i in range(2, len(eoc[key])):
        eoc_v = eoc[key][i]
        if eoc_v < 1.8:
            raise RuntimeError("Order of Convergence < 1.8 detected!")



import pytest
from netgen.geom2d import unit_square
from ngsolve import *
from xfem import *
from xfem.lsetcurv import *
ngsglobals.msg_level = 0

mesh = Mesh(unit_square.GenerateMesh(maxh=0.2))
levelset = (x - 0.5)**2 + (y - 0.5)**2 - 0.33**2


@pytest.mark.parametrize('order', [1, 2, 3])
@pytest.mark.parametrize('DOM', [POS, IF, NEG])
@pytest.mark.parametrize('subdivlvl', [0, 1, 2])
def test_cut_symbols_straightcut(order, DOM, subdivlvl):

    if order > 1 and subdivlvl == 0:
        lsetmeshadap = LevelSetMeshAdaptation(mesh, order=order,
                                              levelset=levelset)
        lset_h = lsetmeshadap.lset_p1
    elif order == 1 and subdivlvl == 0:
        lsetmeshadap = NoDeformation(mesh, levelset)
        lset_h = lsetmeshadap.lset_p1
    elif subdivlvl > 0:
        lsetmeshadap = NoDeformation(mesh, levelset)
        lset_h = GridFunction(H1(mesh, order=order))
        lset_h.Set(levelset)

    deform = lsetmeshadap.deform

    ci = CutInfo(mesh, lset_h)
    els_hasdom = ci.GetElementsOfType(HAS(DOM))
    facets_none = BitArray(mesh.nedge)
    facets_none.Clear()

    V = H1(mesh, order=order)
    u, v = V.TnT()

    gfu = GridFunction(V)
    gfu.Set(sin(x))

    w1, w2 = gfu.vec.CreateVector(), gfu.vec.CreateVector()

    # Bilinear Form
    form = Grad(u) * Grad(v) + u * v

    # Differential Symbol Version
    a1 = RestrictedBilinearForm(V, element_restriction=els_hasdom,
                                facet_restriction=facets_none,
                                check_unused=False)
    a1 += form * dCut(lset_h, DOM, subdivlvl=subdivlvl,
                      definedonelements=els_hasdom, deformation=deform)
    a1.Assemble()
    w1.data = a1.mat * gfu.vec

    # SymbolicBFI version
    lset_dom = {'levelset': lset_h, 'domain_type': DOM, 'subdivlvl': subdivlvl}
    a2 = RestrictedBilinearForm(V, element_restriction=els_hasdom,
                                facet_restriction=facets_none,
                                check_unused=False)
    a2 += SymbolicBFI(levelset_domain=lset_dom, form=form,
                      definedonelements=els_hasdom)
    mesh.SetDeformation(deform)
    a2.Assemble()
    mesh.UnsetDeformation()
    w2.data = a2.mat * gfu.vec

    # Check
    w1.data -= w2
    diff = Norm(w1)
    print(f'diff : {diff}')
    assert diff < 1e-12


@pytest.mark.parametrize('order', [1, 2, 3])
@pytest.mark.parametrize('DOM', [POS, NEG])
@pytest.mark.parametrize('skeleton', [True, False])
@pytest.mark.parametrize('element_boundary', [True, False])
def test_cut_symbols_dg(order, DOM, skeleton, element_boundary):
    if skeleton is False and element_boundary is False:
        return None
    if order > 1:
        lsetmeshadap = LevelSetMeshAdaptation(mesh, order=order,
                                              levelset=levelset)
    else:
        lsetmeshadap = NoDeformation(mesh, levelset)

    lset_h = lsetmeshadap.lset_p1
    deform = lsetmeshadap.deform

    ci = CutInfo(mesh, lset_h)
    els_hasdom = ci.GetElementsOfType(HAS(DOM))
    facets_dom = GetFacetsWithNeighborTypes(mesh, a=els_hasdom, b=els_hasdom)
    els_none = BitArray(mesh.ne)
    els_none[:] = False

    V = H1(mesh, order=order, dgjumps=True)
    u, v = V.TnT()

    gfu = GridFunction(V)
    gfu.Set(sin(x))

    w1, w2 = gfu.vec.CreateVector(), gfu.vec.CreateVector()

    # Bilinear form to integrate
    nF = specialcf.normal(mesh.dim)
    flux_u = -0.5 * (grad(u) + grad(u.Other())) * nF
    flux_v = -0.5 * (grad(v) + grad(v.Other())) * nF
    jump_u = u - u.Other()
    jump_v = v - v.Other()

    form = jump_u * jump_v + flux_u * jump_v + flux_v * jump_u

    # Differential Symbol Version
    a1 = RestrictedBilinearForm(V, element_restriction=els_hasdom,
                                facet_restriction=facets_dom,
                                check_unused=False)
    a1 += form * dCut(lset_h, DOM, definedonelements=facets_dom,
                      deformation=deform, skeleton=skeleton,
                      element_boundary=element_boundary)
    a1.Assemble()
    w1.data = a1.mat * gfu.vec

    # SymbolicBFI version
    lset_dom = {'levelset': lset_h, 'domain_type': DOM, 'subdivlvl': 0}
    a2 = RestrictedBilinearForm(V, element_restriction=els_hasdom,
                                facet_restriction=facets_dom,
                                check_unused=False)
    a2 += SymbolicBFI(levelset_domain=lset_dom, form=form,
                      definedonelements=facets_dom, skeleton=skeleton,
                      element_boundary=element_boundary)
    mesh.SetDeformation(deform)
    a2.Assemble()
    mesh.UnsetDeformation()
    w2.data = a2.mat * gfu.vec

    # Check
    w1.data -= w2
    diff = Norm(w1)
    print(f'diff : {diff}')
    assert diff < 1e-12


@pytest.mark.parametrize('order', [1, 2, 3])
@pytest.mark.parametrize('DOM', [POS, NEG])
def test_cut_symbols_facetpatch(order, DOM):

    # Higher order level set approximation
    if order > 1:
        lsetmeshadap = LevelSetMeshAdaptation(mesh, order=order,
                                              levelset=levelset)
    else:
        lsetmeshadap = NoDeformation(mesh, levelset)

    deform = lsetmeshadap.deform
    lsetp1 = lsetmeshadap.lset_p1

    ci = CutInfo(mesh, lsetp1)
    els_hasdom = ci.GetElementsOfType(HAS(DOM))
    els_if = ci.GetElementsOfType(IF)
    ba_facets = GetFacetsWithNeighborTypes(mesh, a=els_hasdom, b=els_if)
    els_gp = GetElementsWithNeighborFacets(mesh, ba_facets)

    V = H1(mesh, order=order, dgjumps=True)
    u, v = V.TnT()

    gfu = GridFunction(V)
    gfu.Set(sin(x))

    w1, w2 = gfu.vec.CreateVector(), gfu.vec.CreateVector()

    # Bilinear form
    form = (u - u.Other()) * (v - v.Other())

    # Differential Symbol Version
    a1 = RestrictedBilinearForm(V, element_restriction=els_gp,
                                facet_restriction=ba_facets,
                                check_unused=False)
    a1 += form * dFacetPatch(definedonelements=ba_facets, deformation=deform)
    a1.Assemble()
    w1.data = a1.mat * gfu.vec

    # SymbolicBFI version
    a2 = RestrictedBilinearForm(V, element_restriction=els_gp,
                                facet_restriction=ba_facets,
                                check_unused=False)
    a2 += SymbolicFacetPatchBFI(form=form, skeleton=False,
                                definedonelements=ba_facets)
    mesh.SetDeformation(deform)
    a2.Assemble()
    mesh.UnsetDeformation()
    w2.data = a2.mat * gfu.vec

    # Check
    w1.data -= w2
    diff = Norm(w1)
    print(f'diff : {diff}')
    assert diff < 1e-12


@pytest.mark.parametrize('order', [1, 2, 3])
@pytest.mark.parametrize('DOM', [POS, IF, NEG])
@pytest.mark.parametrize('subdivlvl', [0, 1, 2])
def test_cut_symbol_integrate(order, DOM, subdivlvl):

    if order > 1 and subdivlvl == 0:
        lsetmeshadap = LevelSetMeshAdaptation(mesh, order=order,
                                              levelset=levelset)
        lset_h = lsetmeshadap.lset_p1
    elif order == 1 and subdivlvl == 0:
        lsetmeshadap = NoDeformation(mesh, levelset)
        lset_h = lsetmeshadap.lset_p1
    elif subdivlvl > 0:
        lsetmeshadap = NoDeformation(mesh, levelset)
        lset_h = GridFunction(H1(mesh, order=order))
        lset_h.Set(levelset)

    deform = lsetmeshadap.deform

    f = x**2 * sin(y)

    # Differential Symbol version
    dx = dCut(levelset=lset_h, domain_type=DOM, order=order,
              subdivlvl=subdivlvl, deformation=deform)
    integral_dx = Integrate(f * dx, mesh=mesh)

    # Old version
    lset_dom = {'levelset': lset_h, 'domain_type': DOM, 'subdivlvl': subdivlvl}
    mesh.SetDeformation(deform)
    integral = Integrate(levelset_domain=lset_dom, cf=f, mesh=mesh,
                         order=order)
    mesh.UnsetDeformation()

    assert abs(integral_dx - integral) < 1e-12



# solve the Poisson equation -Delta u = f
# with Dirichlet boundary condition u = 0

import pytest
from ngsolve import *
from ngsolve.meshes import *
from netgen.geom2d import unit_square
from xfem import *

ngsglobals.msg_level = 1

@pytest.mark.parametrize("dimension", [2,3])
def test_shifteval(dimension):
  if dimension == 2:
    mesh = MakeStructured2DMesh(quads = False, nx=8, ny=8)
  else:
    mesh = MakeStructured3DMesh(hexes = False, nx=8, ny=8, nz=8)
    
  # H1-conforming finite element space
  fes = H1(mesh, order=3, dirichlet=[1,2,3,4])
  fes_dfm = H1(mesh, order=3, dim=dimension)
  
  gfu_new = GridFunction(fes)
  
  gfu_old = GridFunction(fes)
  
  dfm_back = GridFunction(fes_dfm)
  #dfm_back.vec[37:] = 0.1
  if dimension == 2:
    dfm_back.Set(CoefficientFunction((0.2*sin(5*y),0.2*cos(5*x))))
  else:
    dfm_back.Set(CoefficientFunction((0.15*sin(5*y),0.15*cos(5*z),0.15*sin(5*x))))
    
  for i in range(dimension*mesh.nv):
      dfm_back.vec[i] = 0.0
  dfm_forth = GridFunction(fes_dfm)
  
  mesh.SetDeformation(dfm_back)
  exact = sin(10*y)
  gfu_old.Set(exact)
  l2error_old = sqrt (Integrate ( (gfu_old-exact)*(gfu_old-exact), mesh, order=10))
  mesh.UnsetDeformation()
  Draw (gfu_old,mesh,"gfu_old")
  Draw (dfm_back,mesh,"dfmback")
  gfu_new.Set(shifted_eval(gfu_old,dfm_back,dfm_forth))
  Draw (gfu_new,mesh,"gfu_new")

  error_old = l2error_old
  error_new = sqrt (Integrate ( (gfu_new-exact)*(gfu_new-exact), mesh, order=10))
  print ("L2-error(old):", error_old)
  print ("L2-error(new):", error_new)
  assert error_old < 1e-3
  assert error_new < 1e-3



from netgen.geom2d import SplineGeometry
from netgen.meshing import MeshingParameters
from ngsolve import *
from xfem import *
# For LevelSetAdaptationMachinery
from xfem.lsetcurv import *

r2 = 3/4 # outer radius
r1 = 1/4 # inner radius
rc = (r1+r2) / 2.0
rr = (r2-r1) / 2.0
r = sqrt(x*x+y*y)
levelset = IfPos(r-rc, r-rc-rr,rc-r-rr)

# Geometry 
square = SplineGeometry()
square.AddRectangle([-1,-1],[1,1],bc=1)

def meshonlvl(lvl):
    ngmesh = square.GenerateMesh(maxh=0.1, quad_dominated=False)
    mesh = Mesh (ngmesh)
    for i in range(lvl):
        mesh.Refine()
    return mesh


#trig version:
ngmesh = square.GenerateMesh(maxh=0.1, quad_dominated=False)
mesh = Mesh (ngmesh)
p1prolong = P1Prolongation(mesh)

polorder = 1

order = 1
nref = 2

Vh = H1(mesh, order = polorder, dirichlet=[], dgjumps = True)
lsetp1 = GridFunction(Vh)
InterpolateToP1(levelset,lsetp1)

def GetActiveDof(mesh):
    Vh = H1(mesh, order = polorder, dirichlet=[], dgjumps = True)
    lsetp1 = GridFunction(Vh)
    InterpolateToP1(levelset,lsetp1)
    ci = CutInfo(mesh,lsetp1)
    return GetDofsOfElements(Vh,ci.GetElementsOfType(HASNEG))

Vhc = Compress(Vh,active_dofs=GetActiveDof(mesh))
p1prolong.Update(Vhc)
gfu = GridFunction(Vhc)
print("ndof",Vhc.ndof)
gf = []
gf.append(GridFunction(Compress(H1(mesh,order=polorder),active_dofs=GetActiveDof(mesh))))
for i in range(nref):
    mesh.Refine()
    Vh.Update()
    lsetp1.Update()
    InterpolateToP1(levelset,lsetp1)
    ci = CutInfo(mesh,lsetp1)
    Vhc.SetActiveDofs(GetActiveDof(mesh))
    Vhc.Update()
    print(Vh.ndof)
    print(Vhc.ndof)
    gfu.Update()
    p1prolong.Update(Vhc)

    
    nmesh = meshonlvl(i)
    gf.append(GridFunction(Compress(H1(nmesh,order=polorder),active_dofs=GetActiveDof(nmesh))))
    

    
v = gfu.vec.CreateVector()    
w = gfu.vec.CreateVector()

for i in range(len(v)):
    v[i] = i
w.data = v
# Pv = gfu.vec.CreateVector()    
# Rw = gfu.vec.CreateVector()


for i in range(len(gf[0].vec)):
    gf[0].vec[i] = i


for i in range(nref+1):
    if i > 0:
       p1prolong.Prolongate(i,w) 
    #print(w)
    nmesh = meshonlvl(i)
    gfvis = GridFunction(Compress(H1(nmesh,order=polorder),active_dofs=GetActiveDof(nmesh)))
    for i in range(len(gfvis.vec)):
        gfvis.vec[i] = w[i]
    Draw(gfvis,nmesh,"gf")
    input(i)
    
    
v = gfu.vec.CreateVector()    
w = gfu.vec.CreateVector()

Pv = gfu.vec.CreateVector()    
Rw = gfu.vec.CreateVector()

for i in range(len(v)):
    v[i] = i
for i in range(len(w)):
    w[i] = 1.0/(i+1)

Rw.data = w    
p1prolong.Restrict(2,Rw)
p1prolong.Restrict(1,Rw)

Pv.data = v
p1prolong.Prolongate(1,Pv)
p1prolong.Prolongate(2,Pv)

print(InnerProduct(Pv,w))
print(InnerProduct(v,Rw))
    
# # gfu.vec[:] = 1.0 

# # input("")
# print(gfu.vec)

# p1prolong.Restrict(2,gfu.vec)
# print(gfu.vec)
# input("")
# p1prolong.Restrict(1,gfu.vec)

# print(gfu.vec)
# input("")




"""
	Description:
	------------
	Two disjoint triangles.
"""

from netgen.geom2d import SplineGeometry
from ngsolve import *
from xfem import *
from xfem.mlset import *

geo = SplineGeometry()
geo.AddRectangle((-4, -3), (4, 3), bc=1)

mesh = Mesh(geo.GenerateMesh(maxh=0.5))


level_sets = [x - 3, -x - y + 1, -x + y + 1,
              -x - 3, x - y + 1, x + y + 1 ]

level_sets_p1 = tuple(GridFunction(H1(mesh, order=1))
                      for i in range(len(level_sets)))
for i, lsetp1 in enumerate(level_sets_p1):
    InterpolateToP1(level_sets[i], lsetp1)
    Draw(lsetp1, mesh, "lsetp1_{}".format(i))

triangle1 = DomainTypeArray((NEG, NEG, NEG))
triangle2 = DomainTypeArray((NEG, NEG, NEG))

triangles = DomainTypeArray([(NEG, NEG, NEG, ANY, ANY, ANY),
                            (ANY, ANY, ANY, NEG, NEG, NEG)])


normals1 = triangle1.GetOuterNormals(level_sets_p1[:3])
normals2 = triangle1.GetOuterNormals(level_sets_p1[3:])

for dtt in triangle1.Boundary():
    name = str(dtt)
    for s in ["DOMAIN_TYPE.", ",", " ", "(", ")"]:
        name = name.replace(s, "")
    Draw(normals1[(NEG, NEG, NEG)][dtt], mesh, "normals_"+name)

Draw(triangle1.Indicator(level_sets_p1[:3]), mesh, "triangle1", sd=6)
input("triangle1")

for dtt in triangle2.Boundary():
    name = str(dtt)
    for s in ["DOMAIN_TYPE.", ",", " ", "(", ")"]:
        name = name.replace(s, "")
    Draw(normals2[(NEG, NEG, NEG)][dtt], mesh, "normals_"+name)
Draw(triangle2.Indicator(level_sets_p1[3:]), mesh, "triangle2", sd=6)


# V = H1(mesh,order=1)
# lsetp1_t1_1, lsetp1_t1_2, lsetp1_t1_3 = GridFunction(V), GridFunction(V),\
# 										GridFunction(V)

# lsetp1_t2_1, lsetp1_t2_2, lsetp1_t2_3 = GridFunction(V), GridFunction(V),\
# 										GridFunction(V)

# InterpolateToP1( x     - 3, lsetp1_t1_1)
# InterpolateToP1(-x - y + 1, lsetp1_t1_2)
# InterpolateToP1(-x + y + 1, lsetp1_t1_3)
# InterpolateToP1(-x     - 3, lsetp1_t2_1)
# InterpolateToP1( x - y + 1, lsetp1_t2_2)
# InterpolateToP1( x + y + 1, lsetp1_t2_3)

# Draw(lsetp1_t1_1, mesh, "lsetp1_t1_1")
# Draw(lsetp1_t1_2, mesh, "lsetp1_t1_2")
# Draw(lsetp1_t1_3, mesh, "lsetp1_t1_3")
# Draw(lsetp1_t1_1 * lsetp1_t1_2 * lsetp1_t1_3, mesh, "lsetp1_t1_mult")

# Draw(lsetp1_t2_1, mesh, "lsetp1_t2_1")
# Draw(lsetp1_t2_2, mesh, "lsetp1_t2_2")
# Draw(lsetp1_t2_3, mesh, "lsetp1_t2_3")
# Draw(lsetp1_t2_1 * lsetp1_t2_2 * lsetp1_t2_3, mesh, "lsetp1_t2_mult")


# inner1 = {"levelsets": (lsetp1_t1_1,lsetp1_t1_2,lsetp1_t1_3),
# 		  "domain_type": (NEG,NEG,NEG)
# 		 }

# boundary1 = {"levelsets": (lsetp1_t1_1,lsetp1_t1_2,lsetp1_t1_3)
# 		     "domain_type": (IF,NEG,NEG)|(NEG,IF,NEG)|(NEG,NEG,IF)
# 			}

# outer1 = {"levelsets": (lsetp1_t1_1,lsetp1_t1_2,lsetp1_t1_3),
# 		  "domain_type": ~(NEG,NEG,NEG)
# 		 }

# inner2 = {"levelsets": (lsetp1_t2_1,lsetp1_t2_2,lsetp1_t2_3),
# 		  "domain_type": (NEG,NEG,NEG)
# 		 }

# boundary2 = {"levelsets": (lsetp1_t2_1,lsetp1_t2_2,lsetp1_t2_3)
# 		     "domain_type": (IF,NEG,NEG)|(NEG,IF,NEG)|(NEG,NEG,IF)
# 			}

# outer2 = {"levelsets": (lsetp1_t2_1,lsetp1_t2_2,lsetp1_t2_3),
# 		  "domain_type": ~(NEG,NEG,NEG)
# 		 }

# # Best way to do "outside of both"?

# ouside_of_both = {"levelsets": (lsetp1_t1_1,lsetp1_t1_2,lsetp1_t1_3,
# 								lsetp1_t2_1,lsetp1_t2_2,lsetp1_t2_3),
# 		  		  "domain_type": ~(NEG,NEG,NEG,ANY,ANY,ANY)&~(ANY,ANY,ANY,NEG,NEG,NEG)
# 		 		  }

# # I would prefer something like

# outside_of_both = outer1 & outer2